Package: gdtools
Title: Font Metrics and Font Management Utilities for R Graphics
Version: 0.5.1
Authors@R: c(
    person("David", "Gohel", , "david.gohel@ardata.fr", role = c("aut", "cre")),
    person("Hadley", "Wickham", , "hadley@rstudio.com", role = "aut"),
    person("Lionel", "Henry", , "lionel@rstudio.com", role = "aut"),
    person("Jeroen", "Ooms", , "jeroen@berkeley.edu", role = "aut",
           comment = c(ORCID = "0000-0002-4035-0289")),
    person("Yixuan", "Qiu", role = "ctb"),
    person("R Core Team", role = "cph",
           comment = "Cairo code from X11 device"),
    person("ArData", role = "cph"),
    person("RStudio", role = "cph")
  )
Description: Compute text metrics (width, ascent, descent) using 'Cairo'
    and 'FreeType', independently of the active graphic device. Font
    lookup is delegated to 'systemfonts'. Additional utilities let users
    register 'Google Fonts' or bundled 'Liberation' fonts, check font
    availability, and assemble 'htmlDependency' objects so that fonts are
    correctly embedded in 'Shiny' applications, 'R Markdown' documents,
    and 'htmlwidgets' outputs such as 'ggiraph'.
License: GPL-3 | file LICENSE
URL: https://davidgohel.github.io/gdtools/
BugReports: https://github.com/davidgohel/gdtools/issues
Depends: R (>= 4.0.0)
Imports: fontquiver (>= 0.2.0), htmltools, Rcpp (>= 0.12.12),
        systemfonts (>= 1.3.1), tools
Suggests: curl, gfonts, methods, testthat
LinkingTo: Rcpp
Encoding: UTF-8
SystemRequirements: cairo, freetype2, fontconfig
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-25 16:16:18 UTC; davidgohel
Author: David Gohel [aut, cre],
  Hadley Wickham [aut],
  Lionel Henry [aut],
  Jeroen Ooms [aut] (ORCID: <https://orcid.org/0000-0002-4035-0289>),
  Yixuan Qiu [ctb],
  R Core Team [cph] (Cairo code from X11 device),
  ArData [cph],
  RStudio [cph]
Maintainer: David Gohel <david.gohel@ardata.fr>
Repository: CRAN
Date/Publication: 2026-05-25 17:10:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-25 17:44:01 UTC; unix
Archs: gdtools.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(print,font_set)
export(addGFontHtmlDependency)
export(check_gfonts)
export(dummy_setup)
export(font_family_exists)
export(font_google)
export(font_liberation)
export(font_set)
export(font_set_auto)
export(font_set_liberation)
export(fonts_cache_dir)
export(gfontHtmlDependency)
export(init_fonts_cache)
export(install_gfont_script)
export(installed_gfonts)
export(liberationmonoHtmlDependency)
export(liberationsansHtmlDependency)
export(liberationserifHtmlDependency)
export(m_str_extents)
export(match_family)
export(register_gfont)
export(register_liberationmono)
export(register_liberationsans)
export(register_liberationserif)
export(rm_fonts_cache)
export(set_dummy_conf)
export(str_metrics)
export(strings_sizes)
export(sys_fonts)
export(version_cairo)
export(version_freetype)
importFrom(Rcpp,sourceCpp)
importFrom(fontquiver,font_faces)
importFrom(grDevices,is.raster)
importFrom(htmltools,attachDependencies)
importFrom(htmltools,htmlDependency)
importFrom(htmltools,tags)
importFrom(systemfonts,match_fonts)
importFrom(systemfonts,register_font)
importFrom(systemfonts,registry_fonts)
importFrom(systemfonts,system_fonts)
importFrom(tools,R_user_dir)
importFrom(utils,packageVersion)
useDynLib(gdtools)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# gdtools

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/gdtools)](https://CRAN.R-project.org/package=gdtools)
[![R build
status](https://github.com/davidgohel/gdtools/workflows/R-CMD-check/badge.svg)](https://github.com/davidgohel/gdtools/actions)
[![codecov test
coverage](https://codecov.io/gh/davidgohel/gdtools/branch/main/graph/badge.svg)](https://app.codecov.io/gh/davidgohel/gdtools)
<!-- badges: end -->

`gdtools` addresses four practical problems when working with fonts in
R:

| Problem | gdtools solution |
|----|----|
| Measure text without a device: layout calculations (column widths, text wrapping) need reliable metrics independent of the rendering device | `strings_sizes()` via Cairo |
| Register non-system fonts: Google Fonts or Liberation fonts (bundled) must be made visible to ‘systemfonts’ | `font_set_liberation()`, `font_set_auto()`, `font_set()` |
| Embed fonts in HTML: Shiny, R Markdown and girafe need `htmlDependency` objects so the browser loads the right fonts | `font_set()$dependencies`, `gfontHtmlDependency()`, `liberation*HtmlDependency()` |
| Check font availability: detect missing fonts before rendering, to avoid silent fallbacks | `font_family_exists()`, `sys_fonts()` |

## How font metric calculation works

`gdtools` was originally designed to compute font metrics for R graphic
devices via C-level interfaces (Cairo and FreeType). With the arrival of
the ‘systemfonts’ package, font discovery and matching became much
simpler and the C interface was no longer needed by other packages.

The package has since been refocused. Font metric calculation follows
two steps:

1.  Font lookup: ‘systemfonts’ resolves the family name to a font file.
    Any font registered via `systemfonts::register_font()`,
    `register_gfont()`, or `font_set()` is found.
2.  Metric computation: Cairo reads the font file and computes width,
    ascent, and descent.

The metrics are independent of any graphic device. Whether they match
the final rendering depends on how the device resolves fonts on its
side, this is a device concern.

## Installation

You can install the released version of gdtools from CRAN with:

``` r
install.packages("gdtools")
```

And the development version from GitHub with:

``` r
# install.packages("remotes")
remotes::install_github("davidgohel/gdtools")
```

## Font metrics

The main function for text measurement is `strings_sizes()`. It returns
width, ascent and descent in inches using Cairo. The measurements are
accurate for devices that use Cairo or ‘systemfonts’ for font resolution
(ragg, svglite, ggiraph, `cairo_pdf()`, …); for devices with their own
font engine (`pdf()`, `png()`, …) the metrics may not match the
rendering.

``` r
library(gdtools)
strings_sizes(
  c("a string", "a longer string"),
  fontsize = 24, 
  bold = TRUE, 
  italic = TRUE
)
#>      width    ascent    descent
#> 1 1.203939 0.2413737 0.07259115
#> 2 2.315104 0.2413737 0.07259115
```

All arguments are vectorized:

``` r
strings_sizes(
  c("normal", "bold", "italic", "bold-italic"),
  fontsize = 12,
  bold = c(FALSE, TRUE, FALSE, TRUE),
  italic = c(FALSE, FALSE, TRUE, TRUE)
)
#>       width    ascent     descent
#> 1 0.5094401 0.1195475 0.003173828
#> 2 0.3517253 0.1199544 0.003092448
#> 3 0.3334147 0.1195475 0.002929688
#> 4 0.7870280 0.1206868 0.003092448
```

## Font management

`font_set_auto()` detects the best available system fonts and falls back
to Liberation fonts when needed. It returns a `font_set` object ready to
use with ggplot2, ggiraph, or any ‘systemfonts’-based device:

``` r
library(gdtools)

fonts <- font_set_auto()
fonts
#> Font set
#>   sans:    Arial [system]
#>   serif:   Times New Roman [system]
#>   mono:    Courier New [system]
#>   symbol:  Symbol [system]
#>   4 HTML dependencies
```

For explicit control, `font_set()` lets you pick each role with
`font_google()`, `font_liberation()`, or a plain family name:

``` r
fonts <- font_set(
  sans  = font_google("Open Sans"),
  mono  = font_liberation("mono")
)

fonts$sans          # "Open Sans"
fonts$dsvg_fonts    # named list for dsvg() / girafe()
fonts$dependencies  # list of htmlDependency objects
```

For an offline-only setup, `font_set_liberation()` uses Liberation fonts
for all four roles:

``` r
fonts <- font_set_liberation()
fonts
```

### Going further

#### Google Fonts

A set of lower-level functions is provided to download and cache fonts
from ‘Google Fonts’. They can be used in ‘R Markdown’ documents and
‘Shiny’ applications, as well as with graphic outputs generated by
‘ggiraph’, ‘ragg’ and ‘svglite’, or tabular outputs from ‘flextable’.

``` r
# Download and register with systemfonts (cached for future use)
register_gfont(family = "Open Sans")
```

Use `addGFontHtmlDependency()` to embed a Google Font in an HTML
document, and `gfontHtmlDependency()` to create an `htmltools`
dependency object.

To install a Google Font at the system level (e.g. in a Dockerfile), use
`install_gfont_script()`:

``` r
install_gfont_script("Fira Sans", platform = "debian", file = "install-font.sh")
```

#### Liberation fonts

The package bundles ‘Liberation Sans’, ‘Liberation Serif’ and
‘Liberation Mono’ fonts (SIL Open Font License). They are useful as
cross-platform fallback fonts, for visual tests, or when no internet
connection is available.

`font_set_liberation()` registers all three families with ‘systemfonts’
and produces the matching HTML dependencies in a single call:

``` r
fonts <- font_set_liberation()
fonts$sans
fonts$dependencies
```

The individual functions `register_liberationsans()`,
`register_liberationserif()` and `register_liberationmono()` remain
available when only a specific variant is needed.

## Install from sources on macOS

Please read carefully the official R for macOS instructions:
<https://mac.r-project.org/>

To compile ‘gdtools’ from source, you need `cairo` (which pulls in
`freetype` and `fontconfig` as dependencies) and `gettext` (which
provides `libintl`, required at link time by `fontconfig`). Using
Homebrew is recommended:

    brew install cairo fontconfig gettext

Then configure your `~/.R/Makevars` file with the appropriate paths. For
Apple Silicon Macs (M1/M2/M3/M4), the typical configuration is:

    CFLAGS=-I/opt/homebrew/include
    CPPFLAGS=-I/opt/homebrew/include
    CXXFLAGS=-I/opt/homebrew/include
    CXX11FLAGS=-I/opt/homebrew/include
    LDFLAGS=-L/opt/homebrew/lib

For Intel Macs, paths are usually `/usr/local/include` and
`/usr/local/lib`:

    CFLAGS=-I/usr/local/include
    CPPFLAGS=-I/usr/local/include
    CXXFLAGS=-I/usr/local/include
    CXX11FLAGS=-I/usr/local/include
    LDFLAGS=-L/usr/local/lib

> Note: paths may vary depending on your system configuration. We are
> unable to provide support for installation issues related to macOS
> compilation setup.

#' Compute string extents for a vector of string.
#'
#' For each \code{x} element, determines the width and height of a bounding box that's big enough
#' to (just) enclose the provided text. Unit is pixel.
#' @param x Character vector of strings to measure
#' @param bold,italic Is text bold/italic?. A vector of logical to match with x.
#' @param fontname Font name. A vector of character to match with x.
#' @param fontsize Font size. A vector of numeric to match with x.
#' @param fontfile Font file. A vector of character to match with x.
#' @examples
#' \donttest{
#' # The first run can be slow when font caches are missing
#' # as font files are then being scanned to build those font caches.
#' m_str_extents(letters, fontsize = 1:26)
#' m_str_extents(letters[1:3],
#'   bold = c(TRUE, FALSE, TRUE),
#'   italic = c(FALSE, TRUE, TRUE),
#'   fontname = c("sans", "sans", "sans") )
#' }
#' @export
m_str_extents <- function(x, fontname = "sans", fontsize=10, bold = FALSE, italic = FALSE, fontfile = NULL) {

  stopifnot(is.character(x), is.character(fontname),
            is.numeric(fontsize), is.logical(bold),
            is.logical(italic), is.character(fontfile) || is.null(fontfile))

  max_length <- length(x)
  fontname <- rep(fontname, length.out = max_length)
  fontsize <- rep(fontsize, length.out = max_length)
  bold <- rep(bold, length.out = max_length)
  italic <- rep(italic, length.out = max_length)

  if( is.null(fontfile) )
    fontfile <- rep("", length.out = max_length)
  else
    fontfile <- rep(fontfile, length.out = max_length)

  m_str_extents_(x, fontname, fontsize, bold, italic, fontfile)
}

#' Compute string extents.
#'
#' Determines the width and height of a bounding box that's big enough
#' to (just) enclose the provided text.
#'
#' @param x Character vector of strings to measure
#' @param bold,italic Is text bold/italic?
#' @param fontname Font name
#' @param fontsize Font size
#' @param fontfile Font file
#' @examples
#' str_extents(letters)
#' str_extents("Hello World!", bold = TRUE, italic = FALSE,
#'   fontname = "sans", fontsize = 12)
#' @export
str_extents <- function(x, fontname = "sans", fontsize = 12, bold = FALSE, italic = FALSE, fontfile = "") {
  str_extents_(x = x, fontname = fontname, fontsize = fontsize, bold = bold, italic = italic, fontfile = fontfile)
}

#' Get font metrics for a string.
#'
#' @return A named numeric vector
#' @inheritParams str_extents
#' @examples
#' str_metrics("Hello World!")
#' @export
str_metrics <- function(x, fontname = "sans", fontsize = 12, bold = FALSE, italic = FALSE, fontfile = "") {
  str_metrics_(x = x, fontname = fontname, fontsize = fontsize, bold = bold, italic = italic, fontfile = fontfile)
}

#' Validate glyph entries
#'
#' Determines if strings contain glyphs not part of a font.
#'
#' @param x Character vector of strings
#' @param bold,italic Is text bold/italic?
#' @param fontname Font name
#' @param fontfile Font file
#' @return a logical vector, if a character element is containing at least
#' a glyph that can not be matched in the font table, FALSE is returned.
#'
#' @examples
#' glyphs_match(letters)
#' glyphs_match("\u265E", bold = TRUE)
#' @export
glyphs_match <- function(x, fontname = "sans", bold = FALSE, italic = FALSE, fontfile = "") {
  glyphs_match_(x = x, fontname = fontname, bold = bold, italic = italic, fontfile = fontfile)
}

context("metric info")

test_that("a string has positive dimensions", {
  value <- str_extents("Hello World!")
  expect_true(all( value > 0 ))
})

extents <- function(font, ...) {
  extents <- str_extents("foobar", fontfile = font$ttf, ...)
  as.vector(round(extents, 4))
}


test_that("known fonts have correct metrics", {
  skip_if_not_installed("fontquiver")
  sans <- fontquiver::font("Liberation", "Sans", "Regular")
  mono_bi <- fontquiver::font("Bitstream Vera", "Sans Mono", "Bold Oblique")
  expect_equal(extents(sans), c(34.0254, 8.8281), tolerance = .1)
  expect_equal(extents(mono_bi), c(43.3477, 9.2969), tolerance = .1)
})

test_that("fractional font sizes are correctly measured", {
  skip_if_not_installed("fontquiver")
  sans <- fontquiver::font("Liberation", "Sans", "Regular")
  if (version_freetype() < "2.6.0") {
    skip("Old FreeType return different extents for fractional sizes")
  }
  expect_equal(extents(sans, fontsize = 15.05), c(42.5317, 11.0156), tolerance = .2)
})

liberationsans_to_cache <- function() {
  font_id <- "liberation-sans"

  font_dir_ <- font_dir(id = font_id)
  font_css_ <- css_dir(id = font_id)

  if (!dir.exists(font_dir_) || !dir.exists(font_css_)) {
    dir.create(font_dir_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = list.files(
        system.file(package = "fontLiberation", "fonts/liberation-fonts"),
        full.names = TRUE),
      to = font_dir_, overwrite = TRUE)

    dir.create(font_css_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = system.file(package = "gdtools", "css/liberation-sans.css"),
      to = font_css_, overwrite = TRUE)
  }

}

#' @importFrom fontquiver font_faces
#' @export
#' @title Register font 'Liberation Sans'
#' @description Register font 'Liberation Sans' so that it can be used
#' with devices using the 'systemfonts' package, i.e. the 'flextable'
#' package and graphic outputs generated with the 'ragg', 'svglite'
#' and 'ggiraph' packages.
#' @return TRUE if the operation went ok.
#' @family functions for font management
register_liberationsans <- function() {
  if (!font_family_exists("Liberation Sans")) {
    liberation_lst <- font_faces("Liberation", family = "sans")
    register_font(
      name = "Liberation Sans",
      plain = liberation_lst$plain$ttf,
      bold = liberation_lst$bold$ttf,
      italic = liberation_lst$italic$ttf,
      bolditalic = liberation_lst$bolditalic$ttf
    )
  }
  font_family_exists("Liberation Sans")
}

#' @export
#' @title 'Liberation Sans' Font HTML dependency
#' @description Create an HTML dependency ready
#' to be used in 'Shiny' or 'R Markdown' with
#' 'Liberation Sans' Font.
#' @seealso [gfontHtmlDependency()]
#' @family functions for font management
liberationsansHtmlDependency <- function() {
  pkg_version <- packageVersion("gdtools")
  pkg_version_str <- format(pkg_version)

  font_id <- "liberation-sans"
  liberationsans_to_cache()
  htmlDependency(
    all_files = TRUE,
    name = font_id,
    version = pkg_version_str,
    src = file.path(fonts_cache_dir(), font_id),
    stylesheet = paste0("css/", font_id, ".css")
  )
}

#' @title Google Font specification
#' @description Create a font specification for a Google Font,
#' to be used with [font_set()].
#' @param family family name of a 'Google Fonts', for example,
#' "Open Sans", "Roboto", "Fira Code" or "Fira Sans Condensed".
#' @param subset font subset, a character vector, it defaults to
#' only "latin" and "latin-ext".
#' @return An object of class `font_spec`.
#' @family font set functions
#' @export
#' @examples
#' font_google("Roboto")
font_google <- function(family, subset = c("latin", "latin-ext")) {
  structure(
    list(family = family, subset = subset, source = "google"),
    class = "font_spec"
  )
}

#' @title Liberation Font specification
#' @description Create a font specification for a Liberation font,
#' to be used with [font_set()].
#' @param variant one of `"sans"`, `"serif"`, or `"mono"`.
#' @return An object of class `font_spec`.
#' @family font set functions
#' @export
#' @examples
#' font_liberation("sans")
#' font_liberation("mono")
font_liberation <- function(variant = c("sans", "serif", "mono")) {
  variant <- match.arg(variant)
  families <- c(
    sans = "Liberation Sans",
    serif = "Liberation Serif",
    mono = "Liberation Mono"
  )
  structure(
    list(
      family = families[[variant]],
      variant = variant,
      source = "liberation"
    ),
    class = "font_spec"
  )
}

# CSS format descriptor from file extension
css_font_format <- function(path) {
  ext <- tolower(tools::file_ext(path))
  switch(
    ext,
    ttf = "truetype",
    otf = "opentype",
    woff = "woff",
    woff2 = "woff2",
    "truetype"
  )
}

# Cache a system font and generate its @font-face CSS
sysfont_to_cache <- function(family) {
  font_id <- tolower(gsub("[^a-zA-Z0-9]+", "-", family))
  font_id <- sub("-+$", "", font_id)

  css_file <- css_filepath(id = font_id)
  if (file.exists(css_file)) {
    return(font_id)
  }

  faces <- list(
    regular = list(
      weight = "normal",
      italic = FALSE,
      css_weight = 400L,
      css_style = "normal"
    ),
    bold = list(
      weight = "bold",
      italic = FALSE,
      css_weight = 700L,
      css_style = "normal"
    ),
    italic = list(
      weight = "normal",
      italic = TRUE,
      css_weight = 400L,
      css_style = "italic"
    ),
    bolditalic = list(
      weight = "bold",
      italic = TRUE,
      css_weight = 700L,
      css_style = "italic"
    )
  )

  .font_dir <- font_dir(id = font_id)
  .css_dir <- css_dir(id = font_id)
  dir.create(.font_dir, recursive = TRUE, showWarnings = FALSE)
  dir.create(.css_dir, recursive = TRUE, showWarnings = FALSE)

  css_lines <- character()

  for (face_name in names(faces)) {
    face <- faces[[face_name]]
    matched <- match_fonts(
      family = family,
      weight = face$weight,
      italic = face$italic
    )
    src_path <- matched$path[1L]
    if (is.na(src_path) || !nzchar(src_path) || !file.exists(src_path)) {
      next
    }

    ext <- tolower(tools::file_ext(src_path))
    dest_name <- paste0(font_id, "-", face_name, ".", ext)
    file.copy(src_path, file.path(.font_dir, dest_name), overwrite = TRUE)

    fmt <- css_font_format(src_path)
    css_lines <- c(
      css_lines,
      "@font-face {",
      sprintf("  font-family: '%s';", family),
      sprintf("  font-style: %s;", face$css_style),
      sprintf("  font-weight: %d;", face$css_weight),
      sprintf("  src: url('../fonts/%s') format('%s');", dest_name, fmt),
      "}",
      ""
    )
  }

  writeLines(css_lines, css_file, useBytes = TRUE)
  font_id
}

# Create an htmlDependency for a system font
sysfontHtmlDependency <- function(family) {
  font_id <- sysfont_to_cache(family)
  pkg_version <- packageVersion("gdtools")
  htmlDependency(
    all_files = TRUE,
    name = font_id,
    version = format(pkg_version),
    src = file.path(fonts_cache_dir(), font_id),
    stylesheet = paste0("css/", font_id, ".css")
  )
}

# Process a single font role value into family name + dependency
process_font_role <- function(value) {
  if (is.character(value)) {
    if (!font_family_exists(value)) {
      stop(
        "Font family ",
        shQuote(value),
        " is not available. ",
        "Register it first or use font_google() / font_liberation().",
        call. = FALSE
      )
    }
    dep <- sysfontHtmlDependency(value)
    list(family = value, dependency = dep, source = "system")
  } else if (inherits(value, "font_spec")) {
    if (value$source == "google") {
      check_gfonts(errors = TRUE)
      register_gfont(family = value$family, subset = value$subset)
      dep <- gfontHtmlDependency(family = value$family, subset = value$subset)
      list(family = value$family, dependency = dep, source = "google")
    } else if (value$source == "liberation") {
      register_fn <- switch(
        value$variant,
        sans = register_liberationsans,
        serif = register_liberationserif,
        mono = register_liberationmono
      )
      dep_fn <- switch(
        value$variant,
        sans = liberationsansHtmlDependency,
        serif = liberationserifHtmlDependency,
        mono = liberationmonoHtmlDependency
      )
      register_fn()
      dep <- dep_fn()
      list(family = value$family, dependency = dep, source = "liberation")
    } else {
      stop("Unknown font_spec source: ", shQuote(value$source), call. = FALSE)
    }
  } else {
    stop(
      "Each argument must be a character string, a font_google(), or a font_liberation() object.",
      call. = FALSE
    )
  }
}

#' @title Create a font set
#' @description Bundle font registration, HTML dependencies, and
#' font family names into a single object. Each argument accepts
#' a [font_google()] specification, a [font_liberation()] specification,
#' or a plain character string naming a system font already available.
#'
#' The returned object provides fields ready to use with
#' [ggplot2::theme()] (`$sans`, `$serif`, `$mono`),
#' `girafe(fonts = ...)` or `dsvg(fonts = ...)` (`$dsvg_fonts`),
#' and `girafe(dependencies = ...)` (`$dependencies`).
#' @param sans font for sans-serif text.
#' @param serif font for serif text.
#' @param mono font for monospace text.
#' @param symbol font for symbol text.
#' @return An object of class `font_set` with elements:
#' \describe{
#'   \item{sans}{character, the sans font family name (or NULL)}
#'   \item{serif}{character, the serif font family name (or NULL)}
#'   \item{mono}{character, the mono font family name (or NULL)}
#'   \item{symbol}{character, the symbol font family name (or NULL)}
#'   \item{dependencies}{list of `htmlDependency` objects}
#'   \item{dsvg_fonts}{named list of family names for `dsvg()`}
#'   \item{sources}{named list of source labels}
#' }
#' @family font set functions
#' @seealso [font_set_liberation()], [font_set_auto()]
#' @export
#' @examples
#' fonts <- font_set(sans = font_liberation("sans"))
#' fonts$sans
#' fonts$dsvg_fonts
#' fonts$dependencies
font_set <- function(sans = NULL, serif = NULL, mono = NULL, symbol = NULL) {
  roles <- list(sans = sans, serif = serif, mono = mono, symbol = symbol)
  families <- list()
  dependencies <- list()
  sources <- list()
  dsvg_fonts <- list()

  for (role in names(roles)) {
    value <- roles[[role]]
    if (is.null(value)) {
      next
    }
    result <- process_font_role(value)
    families[[role]] <- result$family
    sources[[role]] <- result$source
    dsvg_fonts[[role]] <- result$family
    if (!is.null(result$dependency)) {
      dependencies <- c(dependencies, list(result$dependency))
    }
  }

  structure(
    list(
      sans = families[["sans"]],
      serif = families[["serif"]],
      mono = families[["mono"]],
      symbol = families[["symbol"]],
      dependencies = dependencies,
      dsvg_fonts = dsvg_fonts,
      sources = sources
    ),
    class = "font_set"
  )
}

#' @title Liberation font set
#' @description Shortcut to create a [font_set()] with all four
#' font roles using Liberation fonts (Sans, Serif, Mono) and
#' Liberation Sans as a fallback for symbols.
#' @return An object of class `font_set`.
#' @family font set functions
#' @seealso [font_set()], [font_set_auto()]
#' @export
#' @examples
#' fonts <- font_set_liberation()
#' fonts
font_set_liberation <- function() {
  font_set(
    sans = font_liberation("sans"),
    serif = font_liberation("serif"),
    mono = font_liberation("mono"),
    symbol = font_liberation("sans")
  )
}

# Try each candidate in order; return the first one found, or NULL.
first_available <- function(candidates) {
  for (f in candidates) {
    if (font_family_exists(f)) {
      return(f)
    }
  }
  NULL
}

#' @title Automatic font set
#' @description Build a [font_set()] by detecting the best available
#' system fonts for each role (sans, serif, mono, symbol).
#'
#' For each role a list of well-known fonts is tried in order.
#' When none is found on the system, the corresponding Liberation
#' font is used as a guaranteed offline fallback.
#'
#' | Role | Candidates (in order) | Fallback |
#' |------|----------------------|----------|
#' | sans | Arial, Helvetica, DejaVu Sans | Liberation Sans |
#' | serif | Times New Roman, Times, DejaVu Serif | Liberation Serif |
#' | mono | Courier New, Courier, DejaVu Sans Mono | Liberation Mono |
#' | symbol | Symbol, Apple Symbols | Liberation Sans |
#'
#' @return An object of class `font_set`.
#' @family font set functions
#' @seealso [font_set()], [font_set_liberation()]
#' @export
#' @examples
#' fonts <- font_set_auto()
#' fonts
font_set_auto <- function() {
  sans_candidates <- c("Arial", "Helvetica", "DejaVu Sans")
  serif_candidates <- c("Times New Roman", "Times", "DejaVu Serif")
  mono_candidates <- c("Courier New", "Courier", "DejaVu Sans Mono")
  symbol_candidates <- c("Symbol", "Apple Symbols")

  sans <- first_available(sans_candidates)
  serif <- first_available(serif_candidates)
  mono <- first_available(mono_candidates)
  symbol <- first_available(symbol_candidates)

  font_set(
    sans = if (!is.null(sans)) sans else font_liberation("sans"),
    serif = if (!is.null(serif)) serif else font_liberation("serif"),
    mono = if (!is.null(mono)) mono else font_liberation("mono"),
    symbol = if (!is.null(symbol)) symbol else font_liberation("sans")
  )
}

#' @export
print.font_set <- function(x, ...) {
  cat("Font set\n")
  roles <- c("sans", "serif", "mono", "symbol")
  for (role in roles) {
    family <- x[[role]]
    if (!is.null(family)) {
      source <- x$sources[[role]]
      cat(sprintf("  %-8s %s [%s]\n", paste0(role, ":"), family, source))
    }
  }
  n_deps <- length(x$dependencies)
  if (n_deps > 0L) {
    cat(sprintf(
      "  %d HTML %s\n",
      n_deps,
      if (n_deps == 1L) "dependency" else "dependencies"
    ))
  }
  invisible(x)
}

test_that("font_liberation() returns font_spec", {
  spec <- font_liberation("sans")
  expect_s3_class(spec, "font_spec")
  expect_identical(spec$family, "Liberation Sans")
  expect_identical(spec$variant, "sans")
  expect_identical(spec$source, "liberation")

  spec_mono <- font_liberation("mono")
  expect_identical(spec_mono$family, "Liberation Mono")
  expect_identical(spec_mono$variant, "mono")

  spec_serif <- font_liberation("serif")
  expect_identical(spec_serif$family, "Liberation Serif")
  expect_identical(spec_serif$variant, "serif")
})

test_that("font_google() returns font_spec", {
  skip_if_not_ready()
  spec <- font_google("Roboto")
  expect_s3_class(spec, "font_spec")
  expect_identical(spec$family, "Roboto")
  expect_identical(spec$subset, c("latin", "latin-ext"))
  expect_identical(spec$source, "google")

  spec2 <- font_google("Open Sans", subset = "latin")
  expect_identical(spec2$subset, "latin")
})

test_that("font_set_liberation() works", {
  fonts <- font_set_liberation()
  expect_s3_class(fonts, "font_set")
  expect_identical(fonts$sans, "Liberation Sans")
  expect_identical(fonts$serif, "Liberation Serif")
  expect_identical(fonts$mono, "Liberation Mono")
  expect_identical(fonts$symbol, "Liberation Sans")
  expect_length(fonts$dependencies, 4L)
  expect_identical(
    fonts$dsvg_fonts,
    list(
      sans = "Liberation Sans",
      serif = "Liberation Serif",
      mono = "Liberation Mono",
      symbol = "Liberation Sans"
    )
  )
  expect_identical(
    fonts$sources,
    list(
      sans = "liberation",
      serif = "liberation",
      mono = "liberation",
      symbol = "liberation"
    )
  )
})

test_that("font_set() with system font string", {
  register_liberationsans()
  fonts <- font_set(sans = "Liberation Sans")
  expect_s3_class(fonts, "font_set")
  expect_identical(fonts$sans, "Liberation Sans")
  expect_null(fonts$serif)
  expect_null(fonts$mono)
  expect_null(fonts$symbol)
  expect_length(fonts$dependencies, 1L)
  expect_s3_class(fonts$dependencies[[1]], "html_dependency")
  expect_identical(fonts$dependencies[[1]]$name, "liberation-sans")
  expect_identical(fonts$sources$sans, "system")
})

test_that("font_set() errors on nonexistent system font", {
  expect_error(
    font_set(sans = "nonexistent_xyz_font_abc"),
    "not available"
  )
})

test_that("font_set() with single liberation font", {
  fonts <- font_set(sans = font_liberation("sans"))
  expect_identical(fonts$sans, "Liberation Sans")
  expect_null(fonts$serif)
  expect_null(fonts$mono)
  expect_null(fonts$symbol)
  expect_length(fonts$dependencies, 1L)
  expect_s3_class(fonts$dependencies[[1]], "html_dependency")
  expect_identical(fonts$dsvg_fonts, list(sans = "Liberation Sans"))
})

test_that("font_set() with mixed liberation fonts", {
  fonts <- font_set(
    sans = font_liberation("sans"),
    mono = font_liberation("mono")
  )
  expect_identical(fonts$sans, "Liberation Sans")
  expect_null(fonts$serif)
  expect_identical(fonts$mono, "Liberation Mono")
  expect_length(fonts$dependencies, 2L)
  expect_identical(
    fonts$dsvg_fonts,
    list(
      sans = "Liberation Sans",
      mono = "Liberation Mono"
    )
  )
})

test_that("font_set() with symbol role", {
  register_liberationsans()
  fonts <- font_set(symbol = "Liberation Sans")
  expect_s3_class(fonts, "font_set")
  expect_null(fonts$sans)
  expect_identical(fonts$symbol, "Liberation Sans")
  expect_identical(fonts$dsvg_fonts, list(symbol = "Liberation Sans"))
  expect_identical(fonts$sources$symbol, "system")
  expect_length(fonts$dependencies, 1L)

  fonts3 <- font_set(
    sans = font_liberation("sans"),
    symbol = font_liberation("mono")
  )
  expect_identical(fonts3$sans, "Liberation Sans")
  expect_identical(fonts3$symbol, "Liberation Mono")
  expect_length(fonts3$dependencies, 2L)
  expect_identical(
    fonts3$dsvg_fonts,
    list(
      sans = "Liberation Sans",
      symbol = "Liberation Mono"
    )
  )
})

test_that("print.font_set() runs without error", {
  fonts <- font_set_liberation()
  expect_output(print(fonts), "Font set")
  expect_output(print(fonts), "Liberation Sans")
  expect_output(print(fonts), "4 HTML dependencies")
})

test_that("system font generates CSS and caches files", {
  register_liberationsans()
  fonts <- font_set(sans = "Liberation Sans")
  dep <- fonts$dependencies[[1]]

  # CSS file exists
  css_path <- file.path(dep$src, "css", paste0(dep$name, ".css"))
  expect_true(file.exists(css_path))
  css_content <- readLines(css_path)
  expect_true(any(grepl("@font-face", css_content)))
  expect_true(any(grepl("Liberation Sans", css_content)))
  expect_true(any(grepl("font-weight: 400", css_content)))
  expect_true(any(grepl("font-weight: 700", css_content)))

  # font files exist
  fonts_path <- file.path(dep$src, "fonts")
  expect_true(dir.exists(fonts_path))
  font_files <- list.files(fonts_path)
  expect_true(length(font_files) >= 1L)
})

test_that("font_set() with google font", {
  skip_if_not_ready()
  skip_if_not(check_gfonts())
  dummy_setup()
  fonts <- font_set(sans = font_google("Open Sans"))
  expect_s3_class(fonts, "font_set")
  expect_identical(fonts$sans, "Open Sans")
  expect_length(fonts$dependencies, 1L)
  expect_identical(fonts$sources$sans, "google")
})

#' @title Check if font family exists.
#'
#' @description Check if a font family exists in available fonts.
#'
#' @return A logical value
#' @param font_family font family name (case sensitive)
#' @param system_only If `TRUE`, only look in system-installed fonts
#'   (as returned by [systemfonts::system_fonts()]). This is useful
#'   to verify that a font will be found by fontconfig, which is the
#'   font resolution mechanism used by Cairo devices at rendering
#'   time. When `FALSE` (the default), both system and registered
#'   fonts are considered.
#' @examples
#' font_family_exists("sans")
#' font_family_exists("Arial")
#' font_family_exists("Courier")
#' font_family_exists("Arial", system_only = TRUE)
#' @export
font_family_exists <- function(font_family = "sans", system_only = FALSE) {
  if (system_only) {
    families <- c(system_fonts()$family, default_sys_fonts()$family)
  } else {
    families <- sys_fonts()$family
  }
  tolower(font_family) %in% tolower(families)
}

#' @title List fonts for 'systemfonts'.
#'
#' @description List system and registry fonts into a data.frame
#' containing columns such as path, family, style, weight and italic.
#' @return A data.frame of font information.
#' @examples
#' sys_fonts()
#' @export
#' @importFrom systemfonts system_fonts registry_fonts
sys_fonts <- function() {
  db_sys <- system_fonts()
  db_reg <- registry_fonts()
  nam <- intersect(colnames(db_sys), colnames(db_reg))
  db_sys <- db_sys[, nam]
  db_reg <- db_reg[, nam]
  font_db <- rbind(db_sys, db_reg, default_sys_fonts())
  font_db
}

default_sys_fonts <- function() {
  dat <- data.frame(
    path = rep(NA_character_, 16),
    index = rep(NA_integer_, 16),
    family = rep(c("sans", "serif", "mono", "symbol"), each = 4),
    style = rep(c("Regular", "Bold", "Italic", "Bold Italic"), 4),
    weight = rep(c("normal", "bold", "normal", "bold"), 4),
    italic = rep(c(FALSE, FALSE, TRUE, TRUE), 4),
    stringsAsFactors = FALSE
  )
  res <- match_fonts(
    family = dat$family,
    weight = dat$weight,
    italic = dat$italic
  )
  dat$path <- res$path
  dat$index <- res$index
  dat
}

#' @importFrom htmltools htmlDependency
#' @importFrom utils packageVersion
#' @export
#' @title 'Google Font' HTML dependency
#' @description Create an HTML dependency ready
#' to be used in 'Shiny' or 'R Markdown'.
#' @details
#' It allows users to use fonts from 'Google Fonts' in an HTML page generated by 'shiny' or 'R Markdown'.
#' At the first request, the font files will be downloaded and stored in a cache on the
#' user's machine, thus avoiding many useless downloads or allowing to work with
#' these fonts afterwards without an Internet connection, in a docker image for example.
#' See [fonts_cache_dir()].
#'
#' The server delivering the font files should not be too busy. That's
#' why a one second pause is added after each download to respect the server's
#' limits. This time can be set with the option `GFONTS_DOWNLOAD_SLEEPTIME` which
#' must be a number of seconds.
#' @param family family name of a 'Google Fonts', for example, "Open Sans", "Roboto",
#' "Fira Code" or "Fira Sans Condensed". Complete list is available with the
#' following command:
#'
#' ```
#' gfonts::get_all_fonts()$family |>
#'   unlist() |>
#'   unique() |>
#'   sort()
#' ```
#' @param subset font subset, a character vector, it defaults to only "latin" and
#' "latin-ext" and can contains values such as "greek", "emoji", "chinese-traditional",
#'
#' Run the following code to get a complete list:
#' ```
#' gfonts::get_all_fonts()$subsets |> unlist() |> unique() |> sort()
#' ```
#'
#' @family functions for font management
#' @return an object defined with [htmltools::htmlDependency()].
#' @examples
#' \dontrun{
#' if (check_gfonts()) {
#'   dummy_setup()
#'   gfontHtmlDependency(family = "Open Sans")
#' }
#' }
gfontHtmlDependency <- function(
  family = "Open Sans",
  subset = c("latin", "latin-ext")
) {
  pkg_version <- packageVersion("gdtools")
  pkg_version_str <- format(pkg_version)

  font_id <- get_font_id(family)
  register_gfont(family = family, subset = subset)

  htmlDependency(
    all_files = TRUE,
    name = font_id,
    version = pkg_version_str,
    src = file.path(fonts_cache_dir(), font_id),
    stylesheet = paste0("css/", font_id, ".css")
  )
}

#' @export
#' @importFrom htmltools tags attachDependencies
#' @title Use a Font in Shiny or Markdown
#' @description Add an empty HTML element attached
#' to an 'HTML Dependency' containing
#' the css and the font files so that the font is available
#' in the HTML page. Multiple families are supported.
#'
#' The htmlDependency is defined with function [gfontHtmlDependency()].
#' @inherit gfontHtmlDependency params details
#' @return an HTML object
#' @family functions for font management
#' @examples
#' \dontrun{
#' if (check_gfonts()) {
#'   dummy_setup()
#'   addGFontHtmlDependency(family = "Open Sans")
#' }
#' }
addGFontHtmlDependency <- function(
  family = "Open Sans",
  subset = c("latin", "latin-ext")
) {
  attachDependencies(
    x = tags$style(""),
    lapply(family, gfontHtmlDependency, subset = subset)
  )
}

get_font_id <- function(family) {
  x <- gfonts_summary()

  if (!family %in% x$family) {
    stop("Font family ", shQuote(family), " has not been found.", call. = FALSE)
  }
  x[x$family %in% family, ]$id
}

#' @importFrom systemfonts register_font
#' @export
#' @title Register a 'Google Font'
#' @description Register a font from 'Google Fonts' so that it can be used
#' with devices using the 'systemfonts' package, i.e. the 'flextable'
#' package and graphic outputs generated with the 'ragg', 'svglite'
#' and 'ggiraph' packages.
#' @inherit gfontHtmlDependency params details
#' @return TRUE if the operation went ok.
#' @family functions for font management
#' @examples
#' \dontrun{
#' if (check_gfonts()) {
#'   dummy_setup()
#'   register_gfont(family = "Roboto")
#' }
#' }
register_gfont <- function(
  family = "Open Sans",
  subset = c("latin", "latin-ext")
) {
  font_id <- get_font_id(family)
  x <- gfonts_summary()
  faces <- reduce_faces(x[x$id %in% font_id, ]$variants[[1]])

  font_to_cache(family = family, faces = faces, subset = subset)

  files <- lapply(faces, function(face) {
    list.files(
      path = font_dir(font_id),
      pattern = paste0(font_id, "(.*)-", face, "\\.ttf"),
      full.names = TRUE
    )
  })

  if (!font_family_exists(family)) {
    files[["name"]] <- family
    do.call(register_font, files)
  }

  font_family_exists(family)
}

#' @export
#' @title Shell command to install a font from 'Google Fonts'
#' @description Create a string containing a system command to execute
#' so that the font from 'Google Fonts' is installed on the system.
#' Its execution may require root permissions, in dockerfile for example.
#' @inherit gfontHtmlDependency params details
#' @param platform "debian" and "windows" and "macos" are supported.
#' @param file script file to generate, optional. If the parameter is
#' specified, a file will be generated ready for execution. If the
#' platform is Windows, administration rights are required to run
#' the script.
#' @return the 'shell' or 'PowerShell' command as a string
#' @family functions for font management
#' @examples
#' \dontrun{
#' if (check_gfonts()) {
#'   dummy_setup()
#'   install_gfont_script(family = "Roboto", platform = "macos")
#' }
#' }
install_gfont_script <- function(
  family = "Open Sans",
  subset = c("latin", "latin-ext"),
  platform = c("debian", "windows", "macos"),
  file = NULL
) {
  platform <- match.arg(platform)
  font_id <- get_font_id(family)
  x <- gfonts_summary()
  faces <- reduce_faces(x[x$id %in% font_id, ]$variants[[1]])

  font_to_cache(family = family, faces = faces, subset = subset)

  if ("debian" %in% platform) {
    str <- debian_sysinstall_command(font_id, dir = "custom-fonts")
  } else if ("windows" %in% platform) {
    str <- windows_sysinstall_command(font_id)
  } else {
    str <- macos_sysinstall_command(
      font_id,
      dir = if (user_cache_exists()) "~/Library/Fonts" else "/Library/Fonts"
    )
  }

  if (!is.null(file)) {
    writeLines(str, file, useBytes = TRUE)
    Sys.chmod(file, mode = "755")
  }

  invisible(str)
}

#' @export
#' @title List installed 'Google Fonts'
#' @description List installed 'Google Fonts' that can be
#' found in the user cache directory.
#' @return families names as a character vector
#' @family functions for font management
#' @examples
#' \dontrun{
#' if (check_gfonts()) {
#'   dummy_setup()
#'   register_gfont(family = "Roboto")
#'   installed_gfonts()
#' }
#' }
installed_gfonts <- function() {
  x <- gfonts_summary()
  fonts_ids <- basename(list.dirs(fonts_cache_dir(), recursive = FALSE))
  x$family[x$id %in% fonts_ids]
}

#' @export
#' @title Checks the operability of 'gfonts'
#' @description Checks that 'gfonts' is installed and can be used.
#' Packages 'curl' and 'gfonts' must be installed and internet
#' connectivity must be active.
#' @param errors if TRUE, function is triggering errors if a condition
#' is not satisfied.
#' @return TRUE or FALSE
#' @keywords internal
check_gfonts <- function(errors = FALSE) {
  has_curl <- requireNamespace("curl", quietly = TRUE)
  has_gfonts <- requireNamespace("gfonts", quietly = TRUE)

  if (!has_curl && errors) {
    stop("package 'curl' is required to download fonts from 'fonts.google'.")
  }
  if (!has_curl) {
    return(FALSE)
  }

  if (!has_gfonts && errors) {
    stop("package 'gfonts' is required to download fonts from 'fonts.google'.")
  }
  if (!has_gfonts) {
    return(FALSE)
  }

  has_connectivity <- curl::has_internet()
  if (!has_connectivity && errors) {
    stop("an internet connection is required to download the font files.")
  }
  if (!has_connectivity) {
    return(FALSE)
  }

  TRUE
}


# utils ----
windows_sysinstall_command <- function(font_id) {
  # https://www.jordanmalcolm.com/deploying-windows-10-fonts-at-scale/
  id_dir <- font_dir(font_id)
  install_cmd <- c(
    sprintf("$FontFolder = \"%s\"", id_dir),
    "$FontItem = Get-Item -Path $FontFolder",
    "$FontList = Get-ChildItem -Path \"$FontItem\\*\" -Include ('*.otf','*.ttc','*.ttf')",
    "",
    "foreach ($Font in $FontList) {",
    "  Copy-Item $Font \"C:\\Windows\\Fonts\"",
    "  New-ItemProperty -Name $Font.BaseName -Path \"HKLM:\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Fonts\" -PropertyType string -Value $Font.name",
    "}",
    ""
  )
  install_cmd <- paste0(install_cmd, collapse = "\n")
  install_cmd
}


debian_sysinstall_command <- function(font_id, dir = "custom-fonts") {
  id_dir <- font_dir(font_id)
  create_dir <- sprintf("mkdir -p /usr/share/fonts/truetype/%s", dir)
  id_dir <- font_dir(font_id)
  install_cmd <- sprintf(
    "find %s -name \"*.ttf\" -exec install -m644 {} /usr/share/fonts/truetype/%s/ \\; || return 1",
    gsub(" ", "\\ ", id_dir, fixed = TRUE),
    dir
  )

  paste(create_dir, install_cmd, "fc-cache -f", sep = ";")
}


macos_sysinstall_command <- function(font_id, dir = "~/Library/Fonts") {
  id_dir <- font_dir(font_id)
  sprintf("cp %s/* %s", gsub(" ", "\\ ", id_dir, fixed = TRUE), dir)
}

ds <- dummy_setup()

test_that("fonts work", {
  skip_if_not_ready()
  skip_if_not(check_gfonts(errors = FALSE))

  fontlist <- gfontHtmlDependency(family = "Open Sans")

  expect_identical(fontlist$name, "open-sans")
  expect_true(grepl("open-sans", fontlist$src[[1]]))
  expect_true(grepl("open-sans.css", fontlist$stylesheet))
  expect_s3_class(fontlist, "html_dependency")

  out <- addGFontHtmlDependency(family = "Roboto")
  expect_s3_class(out, "shiny.tag")
  expect_identical(attr(out, which = "html_dependencies")[[1]]$name, "roboto")
  expect_identical(out$name, "style")

  out <- addGFontHtmlDependency(family = "Open Sans")
  expect_error(addGFontHtmlDependency("notafont"))
  expect_identical(attr(out, which = "html_dependencies")[[1]], fontlist)

  random_font <- gfonts::get_all_fonts()$family[4]
  expect_true(register_gfont(family = random_font))

  expect_identical(
    installed_gfonts(),
    c("Roboto", "Open Sans", random_font)
  )

  flib <- liberationsansHtmlDependency()
  expect_identical(flib$name, "liberation-sans")
  expect_true(grepl("liberation-sans", flib$src[[1]]))
  expect_true(grepl("liberation-sans.css", flib$stylesheet))
  expect_s3_class(flib, "html_dependency")

  expect_true(register_liberationsans())
  expect_true(font_family_exists("Liberation Sans"))

  flib <- liberationserifHtmlDependency()
  expect_identical(flib$name, "liberation-serif")
  expect_true(grepl("liberation-serif", flib$src[[1]]))
  expect_true(grepl("liberation-serif.css", flib$stylesheet))
  expect_s3_class(flib, "html_dependency")

  expect_true(register_liberationserif())
  expect_true(font_family_exists("Liberation Serif"))

  flib <- liberationmonoHtmlDependency()
  expect_identical(flib$name, "liberation-mono")
  expect_true(grepl("liberation-mono", flib$src[[1]]))
  expect_true(grepl("liberation-mono.css", flib$stylesheet))
  expect_s3_class(flib, "html_dependency")

  expect_true(register_liberationmono())
  expect_true(font_family_exists("Liberation Mono"))
})

test_that("system dependent font installation works", {
  skip_if_not(check_gfonts(errors = FALSE))
  si <- Sys.info()
  if (si["sysname"] == "Linux") {
    sysnm <- "debian"
  } else if (si["sysname"] == "Darwin") {
    sysnm <- "macos"
  } else if (si["sysname"] == "macOS") {
    sysnm <- "macos"
  } else if (si["sysname"] == "Windows") {
    sysnm <- "windows"
  } else {
    sysnm <- NULL
  }
  if (!is.null(sysnm)) {
    expect_silent(
      command <- install_gfont_script(family = "Roboto", platform = sysnm)
    )
    expect_silent(install_gfont_script(file = tempfile()))
    # Check some part of the command -> Windows is different
    if (si["sysname"] %in% c("Linux", "macOS", "Darwin")) {
      expect_true(grepl(paste0(ds, "/roboto/fonts"), command))
    }
  } else {
    skip(
      "Skipping install_gfont_script test because the platform is not supported"
    )
  }

  # all platform install commands run without error
  expect_no_error(gdtools:::windows_sysinstall_command("roboto"))
  expect_no_error(gdtools:::debian_sysinstall_command("roboto"))
  expect_no_error(gdtools:::macos_sysinstall_command("roboto"))
})

test_that("font-caching works", {
  expect_warning(ds <- dummy_setup())

  expect_silent(fcd <- fonts_cache_dir())
  expect_true(dir.exists(fcd))
  expect_true(grepl("GDTOOLS_CACHE_DIR", fcd))

  options(GDTOOLS_CACHE_DIR = tempdir())
  expect_silent(fcd <- fonts_cache_dir())
  expect_true(dir.exists(fcd))
  expect_false(grepl("GDTOOLS_CACHE_DIR", fcd))

  options(GDTOOLS_CACHE_DIR = NULL)

  Sys.setenv(GDTOOLS_CACHE_DIR = "")

  init_fonts_cache()
  expect_silent(fcd <- fonts_cache_dir())
  expect_true(dir.exists(fcd))
  expect_true(grepl("gdtools", fcd))

  rm_fonts_cache()
  expect_false(dir.exists(fonts_cache_dir()))
})

#' Find best family match with systemfonts
#'
#' \code{match_family()} returns the best font family match.
#'
#' @param font family or face to match.
#' @param bold Whether to match a font featuring a \code{bold} face.
#' @param italic Whether to match a font featuring an \code{italic} face.
#' @param debug deprecated
#' @export
#' @examples
#' match_family("sans")
#' match_family("serif")
#' @importFrom systemfonts match_fonts system_fonts
#' @keywords internal
match_family <- function(
  font = "sans",
  bold = TRUE,
  italic = TRUE,
  debug = NULL
) {
  if (bold) {
    weight <- "bold"
  } else {
    weight <- "normal"
  }
  font <- match_fonts(font, weight = weight, italic = italic)

  sysfonts <- sys_fonts()
  match <- which(sysfonts$path %in% font$path)
  sysfonts$family[match[1]]
}


#' Set and unset a minimalistic Fontconfig configuration
#'
#' @note
#' Fontconfig is not used anymore and these functions will be deprecated
#' in the next release.
#' @export
#' @keywords internal
set_dummy_conf <- function() {
  .Deprecated(msg = "The function has no effect and should not be used.")
  invisible(FALSE)
}

test_that("generic utils work", {
  ve_c <- version_cairo()

  expect_true(ve_c > numeric_version('0'))
  expect_s3_class(ve_c, "package_version")
  expect_s3_class(ve_c, "numeric_version")

  ve_ft <- version_freetype()

  expect_true(ve_ft > numeric_version('0'))
  expect_s3_class(ve_ft, "package_version")
  expect_s3_class(ve_ft, "numeric_version")

  skip_on_cran()
  font_set_liberation()
  expect_true(is.character(match_family("serif")))
  expect_true(is.character(match_family("sans", bold = FALSE, italic = FALSE)))
})

test_that("set_dummy_conf is deprecated", {
  expect_warning(result <- set_dummy_conf(), "should not be used")
  expect_false(result)
})

if(!file.exists("../windows/cairo/include/cairo/cairo.h")){
  unlink("../windows", recursive = TRUE)
  url <- if(grepl("aarch", R.version$platform)){
    "https://github.com/r-windows/bundles/releases/download/cairo-1.18.0/cairo-1.18.0-clang-aarch64.tar.xz"
  } else if(grepl("clang", Sys.getenv('R_COMPILED_BY'))){
    "https://github.com/r-windows/bundles/releases/download/cairo-1.18.0/cairo-1.18.0-clang-x86_64.tar.xz"
  }  else if(getRversion() >= "4.2") {
    "https://github.com/r-windows/bundles/releases/download/cairo-1.18.0/cairo-1.18.0-ucrt-x86_64.tar.xz"
  } else {
    "https://github.com/rwinlib/cairo/archive/v1.16.0.tar.gz"
  }
  download.file(url, basename(url), quiet = TRUE)
  dir.create("../windows", showWarnings = FALSE)
  untar(basename(url), exdir = "../windows", tar = 'internal')
  unlink(basename(url))
  setwd("../windows")
  file.rename(list.files(), 'cairo')
}

# Liberation Sans -----
liberationsans_to_cache <- function() {
  font_id <- "liberation-sans"

  font_dir_ <- font_dir(id = font_id)
  font_css_ <- css_dir(id = font_id)

  if (!dir.exists(font_dir_) || !dir.exists(font_css_)) {
    dir.create(font_dir_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = list.files(
        system.file(package = "fontLiberation", "fonts/liberation-fonts"),
        full.names = TRUE
      ),
      to = font_dir_,
      overwrite = TRUE
    )

    dir.create(font_css_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = system.file(package = "gdtools", "css/liberation-sans.css"),
      to = font_css_,
      overwrite = TRUE
    )
  }
}

#' @importFrom fontquiver font_faces
#' @export
#' @title Register font 'Liberation Sans'
#' @description Register font 'Liberation Sans' so that it can be used
#' with devices using the 'systemfonts' package, i.e. the 'flextable'
#' package and graphic outputs generated with the 'ragg', 'svglite'
#' and 'ggiraph' packages.
#' @param name the name to use for the font family when
#' registering with 'systemfonts'. Using a custom name
#' (e.g. `"sans"`, `"serif"`, `"mono"`) allows devices like
#' 'ragg' to resolve generic family names to this font.
#' @return TRUE if the operation went ok.
#' @family functions for font management
#' @seealso [font_set_liberation()], [font_set()]
#' @examples
#' register_liberationsans()
#' register_liberationsans(name = "sans")
register_liberationsans <- function(name = "Liberation Sans") {
  if (!font_family_exists(name)) {
    liberation_lst <- font_faces("Liberation", family = "sans")
    register_font(
      name = name,
      plain = liberation_lst$plain$ttf,
      bold = liberation_lst$bold$ttf,
      italic = liberation_lst$italic$ttf,
      bolditalic = liberation_lst$bolditalic$ttf
    )
  }
  font_family_exists(name)
}

#' @export
#' @title 'Liberation Sans' Font HTML dependency
#' @description Create an HTML dependency ready
#' to be used in 'Shiny' or 'R Markdown' with
#' 'Liberation Sans' Font.
#' @seealso [font_set_liberation()], [font_set()], [gfontHtmlDependency()]
#' @family functions for font management
liberationsansHtmlDependency <- function() {
  pkg_version <- packageVersion("gdtools")
  pkg_version_str <- format(pkg_version)

  font_id <- "liberation-sans"
  liberationsans_to_cache()
  htmlDependency(
    all_files = TRUE,
    name = font_id,
    version = pkg_version_str,
    src = file.path(fonts_cache_dir(), font_id),
    stylesheet = paste0("css/", font_id, ".css")
  )
}


# Liberation Serif -----
liberationserif_to_cache <- function() {
  font_id <- "liberation-serif"

  font_dir_ <- font_dir(id = font_id)
  font_css_ <- css_dir(id = font_id)

  if (!dir.exists(font_dir_) || !dir.exists(font_css_)) {
    dir.create(font_dir_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = list.files(
        system.file(package = "fontLiberation", "fonts/liberation-fonts"),
        full.names = TRUE
      ),
      to = font_dir_,
      overwrite = TRUE
    )

    dir.create(font_css_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = system.file(package = "gdtools", "css/liberation-serif.css"),
      to = font_css_,
      overwrite = TRUE
    )
  }
}


#' @export
#' @title Register font 'Liberation Serif'
#' @description Register font 'Liberation Serif' so that it can be used
#' with devices using the 'systemfonts' package, i.e. the 'flextable'
#' package and graphic outputs generated with the 'ragg', 'svglite'
#' and 'ggiraph' packages.
#' @inheritParams register_liberationsans
#' @return TRUE if the operation went ok.
#' @family functions for font management
#' @seealso [font_set_liberation()], [font_set()]
#' @examples
#' register_liberationserif()
#' register_liberationserif(name = "serif")
register_liberationserif <- function(name = "Liberation Serif") {
  if (!font_family_exists(name)) {
    liberation_lst <- font_faces("Liberation", family = "serif")
    register_font(
      name = name,
      plain = liberation_lst$plain$ttf,
      bold = liberation_lst$bold$ttf,
      italic = liberation_lst$italic$ttf,
      bolditalic = liberation_lst$bolditalic$ttf
    )
  }
  font_family_exists(name)
}

#' @export
#' @title 'Liberation Serif' Font HTML dependency
#' @description Create an HTML dependency ready
#' to be used in 'Shiny' or 'R Markdown' with
#' 'Liberation Serif' Font.
#' @seealso [font_set_liberation()], [font_set()], [gfontHtmlDependency()]
#' @family functions for font management
liberationserifHtmlDependency <- function() {
  pkg_version <- packageVersion("gdtools")
  pkg_version_str <- format(pkg_version)

  font_id <- "liberation-serif"
  liberationserif_to_cache()
  htmlDependency(
    all_files = TRUE,
    name = font_id,
    version = pkg_version_str,
    src = file.path(fonts_cache_dir(), font_id),
    stylesheet = paste0("css/", font_id, ".css")
  )
}

# Liberation Mono -----
liberationmono_to_cache <- function() {
  font_id <- "liberation-mono"

  font_dir_ <- font_dir(id = font_id)
  font_css_ <- css_dir(id = font_id)

  if (!dir.exists(font_dir_) || !dir.exists(font_css_)) {
    dir.create(font_dir_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = list.files(
        system.file(package = "fontLiberation", "fonts/liberation-fonts"),
        full.names = TRUE
      ),
      to = font_dir_,
      overwrite = TRUE
    )

    dir.create(font_css_, showWarnings = FALSE, recursive = TRUE)
    file.copy(
      from = system.file(package = "gdtools", "css/liberation-mono.css"),
      to = font_css_,
      overwrite = TRUE
    )
  }
}


#' @export
#' @title Register font 'Liberation Mono'
#' @description Register font 'Liberation Mono' so that it can be used
#' with devices using the 'systemfonts' package, i.e. the 'flextable'
#' package and graphic outputs generated with the 'ragg', 'svglite'
#' and 'ggiraph' packages.
#' @inheritParams register_liberationsans
#' @return TRUE if the operation went ok.
#' @family functions for font management
#' @seealso [font_set_liberation()], [font_set()]
#' @examples
#' register_liberationmono()
#' register_liberationmono(name = "mono")
register_liberationmono <- function(name = "Liberation Mono") {
  if (!font_family_exists(name)) {
    liberation_lst <- font_faces("Liberation", family = "mono")
    register_font(
      name = name,
      plain = liberation_lst$plain$ttf,
      bold = liberation_lst$bold$ttf,
      italic = liberation_lst$italic$ttf,
      bolditalic = liberation_lst$bolditalic$ttf
    )
  }
  font_family_exists(name)
}

#' @export
#' @title 'Liberation Mono' Font HTML dependency
#' @description Create an HTML dependency ready
#' to be used in 'Shiny' or 'R Markdown' with
#' 'Liberation Mono' Font.
#' @seealso [font_set_liberation()], [font_set()], [gfontHtmlDependency()]
#' @family functions for font management
liberationmonoHtmlDependency <- function() {
  pkg_version <- packageVersion("gdtools")
  pkg_version_str <- format(pkg_version)

  font_id <- "liberation-mono"
  liberationmono_to_cache()
  htmlDependency(
    all_files = TRUE,
    name = font_id,
    version = pkg_version_str,
    src = file.path(fonts_cache_dir(), font_id),
    stylesheet = paste0("css/", font_id, ".css")
  )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

str_metrics_ <- function(x, fontname = "sans", fontsize = 12, bold = FALSE, italic = FALSE, fontfile = "") {
    .Call('_gdtools_str_metrics_', PACKAGE = 'gdtools', x, fontname, fontsize, bold, italic, fontfile)
}

m_str_extents_ <- function(x, fontname, fontsize, bold, italic, fontfile) {
    .Call('_gdtools_m_str_extents_', PACKAGE = 'gdtools', x, fontname, fontsize, bold, italic, fontfile)
}

m_str_metrics_ <- function(x, fontname, fontsize, bold, italic, fontfile) {
    .Call('_gdtools_m_str_metrics_', PACKAGE = 'gdtools', x, fontname, fontsize, bold, italic, fontfile)
}

#' @rdname version_cairo
#' @export
version_freetype <- function() {
    .Call('_gdtools_version_freetype', PACKAGE = 'gdtools')
}

version_cairo_ <- function() {
    .Call('_gdtools_version_cairo_', PACKAGE = 'gdtools')
}

#' Draw/preview a raster into a string
#'
#' \code{raster_view} is a helper function for testing. It uses \code{htmltools}
#' to render a png as an image with base64 encoded data image.
#'
#' @param x A raster object
#' @param width,height Width and height in pixels.
#' @param interpolate A logical value indicating whether to linearly
#'   interpolate the image.
#' @param code base64 code of a raster
#' @export
#' @examples
#' r <- as.raster(matrix(hcl(0, 80, seq(50, 80, 10)),
#'  nrow = 4, ncol = 5))
#' code <- raster_str(r, width = 50, height = 50)
#' if (interactive() && require("htmltools")) {
#'   raster_view(code = code)
#' }
raster_str <- function(x, width = 480, height = 480, interpolate = FALSE) {

  stopifnot(is.raster(x))

  value <- base64_raster_encode(x,
                w = ncol(x), h = nrow(x), width = width, height = height,
                interpolate = interpolate)
  value
}


#' @export
#' @rdname raster_str
raster_view <- function(code) {
  img <- paste0("<img src=\"data:image/png;base64,", code, "\" />")
  htmltools::browsable(htmltools::HTML(img))
}




#' Draw/preview a raster to a png file
#'
#' @param x A raster object
#' @param path name of the file to create
#' @param width,height Width and height in pixels.
#' @param interpolate A logical value indicating whether to linearly
#'   interpolate the image.
#' @export
#' @examples
#' r <- as.raster(matrix(hcl(0, 80, seq(50, 80, 10)),
#'  nrow = 4, ncol = 5))
#' filepng <- tempfile(fileext = ".png")
#' raster_write(x = r, path = filepng, width = 50, height = 50)
raster_write <- function(x, path, width = 480, height = 480,
                         interpolate = FALSE) {
  stopifnot(is.raster(x))

  raster_png_(raster_ = x,
      w = ncol(x), h = nrow(x),
      width = as.double(width), height = as.double(height),
      interpolate = interpolate,
      filename = path)

  invisible(path)
}

context("raster")

test_that("raster_str", {
  r <- as.raster(matrix(hcl(0, 80, seq(40, 80, 10)), nrow = 5, ncol = 4))
  code <- raster_str(r, width = 50, height = 50)
  expect_gt(nchar(x = code), 50)
})

test_that("raster_write", {
  r <- as.raster(matrix(hcl(0, 80, seq(40, 80, 10)), nrow = 5, ncol = 4))
  filename <- tempfile(fileext = ".png")
  raster_write(r, path = filename, width = 50, height = 50)
  expect_true(file.exists(filename))
})

#' @title Get font metrics for a string.
#'
#' @description Compute the width, ascent and descent of a single
#' string in points using Cairo text measurement.
#'
#' @note This function will be deprecated in a future release.
#' Use [strings_sizes()] instead.
#'
#' @return A named numeric vector with elements \code{width},
#' \code{ascent} and \code{descent} (unit is points).
#' @param x Character vector of strings to measure (only the first element is used).
#' @param bold,italic Is text bold/italic?
#' @param fontname Font name.
#' @param fontsize Font size in points.
#' @param fontfile Path to a font file. If empty, the font is
#' resolved by name via \code{systemfonts}.
#' @family functions for font metrics
#' @examples
#' str_metrics("Hello World!")
#' @export
#' @keywords internal
str_metrics <- function(
  x,
  fontname = "sans",
  fontsize = 12,
  bold = FALSE,
  italic = FALSE,
  fontfile = ""
) {
  str_metrics_(
    x = x,
    fontname = fontname,
    fontsize = fontsize,
    bold = bold,
    italic = italic,
    fontfile = fontfile
  )
}

#' Compute string extents for a vector of strings.
#'
#' For each \code{x} element, determines the width and height of a bounding box that's big enough
#' to (just) enclose the provided text. Unit is points.
#'
#' @note This function will be deprecated in a future release.
#' Use [strings_sizes()] instead.
#' @param x Character vector of strings to measure
#' @param bold,italic Is text bold/italic? A logical vector, recycled to match \code{x}.
#' @param fontname Font name. A character vector, recycled to match \code{x}.
#' @param fontsize Font size. A numeric vector, recycled to match \code{x}.
#' @param fontfile Font file. A character vector, recycled to match \code{x}.
#' @examples
#' \donttest{
#' # The first run can be slow when font caches are missing
#' # as font files are then being scanned to build those font caches.
#' m_str_extents(letters, fontsize = 1:26)
#' m_str_extents(letters[1:3],
#'   bold = c(TRUE, FALSE, TRUE),
#'   italic = c(FALSE, TRUE, TRUE),
#'   fontname = c("sans", "sans", "sans") )
#' }
#' @family functions for font metrics
#' @export
#' @keywords internal
m_str_extents <- function(
  x,
  fontname = "sans",
  fontsize = 10,
  bold = FALSE,
  italic = FALSE,
  fontfile = NULL
) {
  stopifnot(
    is.character(x),
    is.character(fontname),
    is.numeric(fontsize),
    is.logical(bold),
    is.logical(italic),
    is.character(fontfile) || is.null(fontfile)
  )

  max_length <- length(x)
  fontname <- rep(fontname, length.out = max_length)
  fontsize <- rep(fontsize, length.out = max_length)
  bold <- rep(bold, length.out = max_length)
  italic <- rep(italic, length.out = max_length)

  if (is.null(fontfile)) {
    fontfile <- rep("", length.out = max_length)
  } else {
    fontfile <- rep(fontfile, length.out = max_length)
  }

  m_str_extents_(x, fontname, fontsize, bold, italic, fontfile)
}


#' @title Compute strings sizes
#' @description
#' Determines widths, ascent and descent in inches.
#' Font lookup is performed by 'systemfonts' (so any font
#' registered via [systemfonts::register_font()],
#' [register_gfont()], or [font_set()] is found), then Cairo
#' computes the actual metrics. The results are accurate for
#' devices whose rendering finds the same font -- this is
#' guaranteed for 'systemfonts'-based devices (ragg, svglite,
#' ggiraph) and true for Cairo devices (`cairo_pdf()`, ...) when
#' the font is also installed at the system level. For devices
#' with their own font engine (`pdf()`, `png()`, ...) the
#' metrics may not match the rendering.
#'
#' @param x A character vector of strings to measure. All arguments are vectorized
#'   and recycled to match the length of \code{x}.
#' @param fontname A character vector specifying the font family name (e.g., "sans", "serif", "mono").
#'   Default is "sans". This argument is vectorized.
#' @param fontsize A numeric vector specifying the font size in points.
#'   Default is 10. This argument is vectorized.
#' @param bold A logical vector indicating whether the text should be bold.
#'   Default is FALSE. This argument is vectorized.
#' @param italic A logical vector indicating whether the text should be italic.
#'   Default is FALSE. This argument is vectorized.
#' @examples
#' strings_sizes(letters)
#' strings_sizes("Hello World!", bold = TRUE, italic = FALSE,
#'   fontname = "sans", fontsize = 12)
#' @family functions for font metrics
#' @export
strings_sizes <- function(
  x,
  fontname = "sans",
  fontsize = 10,
  bold = FALSE,
  italic = FALSE
) {
  stopifnot(
    is.character(x),
    is.character(fontname),
    is.numeric(fontsize),
    is.logical(bold),
    is.logical(italic)
  )

  max_length <- length(x)
  fontname <- rep(fontname, length.out = max_length)
  fontsize <- rep(fontsize, length.out = max_length)
  bold <- rep(bold, length.out = max_length)
  italic <- rep(italic, length.out = max_length)

  z <- m_str_metrics_(
    x,
    fontname,
    fontsize,
    as.integer(bold),
    as.integer(italic),
    rep("", max_length)
  )
  z$width <- z$width / 72
  z$ascent <- z$ascent / 72
  z$descent <- z$descent / 72
  z
}

context("font file metric info")

test_that("a string has positive dimensions", {
  skip_if_not_installed("fontquiver")
  file <- fontquiver::font("Bitstream Vera", "Sans", "Bold")$ttf
  value <- str_metrics(x = "Hello World", fontsize = 12, fontfile = file)
  expect_true(value["width"] > 0)
  expect_true(value["ascent"] > 0)
  expect_false(value["descent"] < 0)
})

test_that("strings_sizes face mapping is correct", {
  # italic must use face 3, not face 4 (bold-italic)
  plain <- strings_sizes("Test", fontname = "sans", fontsize = 12)
  bold <- strings_sizes("Test", fontname = "sans", fontsize = 12, bold = TRUE)
  italic <- strings_sizes(
    "Test",
    fontname = "sans",
    fontsize = 12,
    italic = TRUE
  )
  bi <- strings_sizes(
    "Test",
    fontname = "sans",
    fontsize = 12,
    bold = TRUE,
    italic = TRUE
  )

  # italic and bold-italic must differ
  expect_false(italic$width == bi$width)
  # all faces must return positive widths
  expect_true(all(c(plain$width, bold$width, italic$width, bi$width) > 0))
})

test_that("strings_sizes vectorised faces are correct", {
  res <- strings_sizes(
    rep("Test", 4),
    fontname = "sans",
    fontsize = 12,
    bold = c(FALSE, TRUE, FALSE, TRUE),
    italic = c(FALSE, FALSE, TRUE, TRUE)
  )
  # italic (row 3) and bold-italic (row 4) must differ
  expect_false(res$width[3] == res$width[4])
})

test_that("NA strings produce NA metrics", {
  res <- strings_sizes(c("hello", NA_character_, "world"))
  expect_true(is.na(res$width[2]))
  expect_true(is.na(res$ascent[2]))
  expect_true(is.na(res$descent[2]))
  expect_true(res$width[1] > 0)
  expect_true(res$width[3] > 0)

  mstre <- m_str_extents(c("a", NA_character_), fontsize = 12)
  expect_true(is.na(mstre[2, 1]))
  expect_true(is.na(mstre[2, 2]))
  expect_true(mstre[1, 1] > 0)
})

test_that("m_str_extents works ", {
  mstre <- m_str_extents(letters, fontsize = 1:26)
  expect_equal(dim(mstre), c(26, 2))
  expect_true(all(is.numeric(mstre)))

  mstre <- m_str_extents(
    letters[1:3],
    bold = c(TRUE, FALSE, TRUE),
    italic = c(FALSE, TRUE, TRUE),
    fontname = c("sans", "sans", "sans")
  )
  expect_equal(dim(mstre), c(3, 2))
  expect_true(all(is.numeric(mstre)))
})

#' @importFrom tools R_user_dir
#' @export
#' @title Manage Font Working Directory
#' @description Initialize or remove font directory used
#' to store downloaded font files.
#'
#' This directory is managed by R function [R_user_dir()] but can also
#' be defined in a non-user location by setting ENV variable `GDTOOLS_CACHE_DIR`
#' or by setting R option `GDTOOLS_CACHE_DIR`.
#'
#' Its value can be read with the `fonts_cache_dir()` function.
#'
#' The directory can be deleted with `rm_fonts_cache()` and
#' created with `init_fonts_cache()`.
#' @family functions for font management
#' @examples
#' fonts_cache_dir()
#'
#' options(GDTOOLS_CACHE_DIR = tempdir())
#' fonts_cache_dir()
#' options(GDTOOLS_CACHE_DIR = NULL)
#'
#' Sys.setenv(GDTOOLS_CACHE_DIR = tempdir())
#' fonts_cache_dir()
#' Sys.setenv(GDTOOLS_CACHE_DIR = "")
#'
#'
#'
#' init_fonts_cache()
#' dir.exists(fonts_cache_dir())
#'
#' rm_fonts_cache()
#' dir.exists(fonts_cache_dir())
fonts_cache_dir <- function() {
  if (dir.exists(Sys.getenv("GDTOOLS_CACHE_DIR"))) {
    dir <- Sys.getenv("GDTOOLS_CACHE_DIR")
  } else if (
    !is.null(getOption("GDTOOLS_CACHE_DIR")) &&
      dir.exists(getOption("GDTOOLS_CACHE_DIR"))
  ) {
    dir <- getOption("GDTOOLS_CACHE_DIR")
  } else {
    dir <- R_user_dir(package = "gdtools", which = "data")
  }
  if (is.null(getOption("GFONTS_DOWNLOAD_SLEEPTIME"))) {
    options(GFONTS_DOWNLOAD_SLEEPTIME = 1)
  }

  dir
}

#' @title Set up a temporary font cache
#' @description Create a temporary directory used as the
#' 'Google Fonts' cache. This is intended for examples and tests.
#' @return The path to the temporary cache directory (invisibly).
#' @export
#' @keywords internal
dummy_setup <- function() {
  options(GDTOOLS_CACHE_DIR = file.path(tempdir(), "GDTOOLS_CACHE_DIR"))
  dir.create(
    getOption("GDTOOLS_CACHE_DIR"),
    recursive = TRUE,
    showWarnings = TRUE
  )
  fonts_cache_dir()
}


user_cache_exists <- function() {
  fonts_cache_dir() == R_user_dir(package = "gdtools", which = "data")
}

#' @export
#' @rdname fonts_cache_dir
rm_fonts_cache <- function() {
  dir <- fonts_cache_dir()
  unlink(dir, recursive = TRUE, force = TRUE)
}


#' @export
#' @rdname fonts_cache_dir
init_fonts_cache <- function() {
  rm_fonts_cache()
  dir.create(fonts_cache_dir(), showWarnings = FALSE, recursive = TRUE)
  liberationsans_to_cache()
  fonts_cache_dir()
}

css_filepath <- function(id) {
  file.path(css_dir(id = id), paste0(id, ".css"))
}

css_dir <- function(id) {
  file.path(fonts_cache_dir(), id, "css")
}
font_dir <- function(id) {
  file.path(fonts_cache_dir(), id, "fonts")
}
reduce_faces <- function(variants) {
  faces <- c(
    plain = "regular",
    italic = "italic",
    bold = "700",
    bolditalic = "700italic"
  )
  faces <- faces[faces %in% variants]
  faces <- faces[!is.na(faces)]
  faces
}

font_to_cache <- function(
  family,
  faces = NULL,
  subset = c("latin", "latin-ext")
) {
  stopifnot(
    `'family' is expected to be a character value` = is.character(family),
    `family is expected to be a single value` = length(family) == 1L
  )

  x <- gfonts_summary()

  if (!family %in% x$family) {
    stop(
      "family ",
      shQuote(family),
      " is not in the fonts provided by 'google fonts'."
    )
  }

  font_id <- x[x$family %in% family, ]$id

  css_file <- css_filepath(id = font_id)
  if (file.exists(css_file)) {
    return(TRUE)
  }

  check_gfonts(errors = TRUE)

  if (is.null(faces)) {
    faces <- reduce_faces(x[x$id %in% font_id, ]$variants[[1]])
  }

  subsets <- intersect(x[x$id %in% font_id, ]$subsets[[1]], subset)

  .font_dir <- font_dir(id = font_id)
  .css_dir <- css_dir(id = font_id)

  dir.create(.font_dir, recursive = TRUE, showWarnings = FALSE)
  dir.create(.css_dir, recursive = TRUE, showWarnings = FALSE)

  gfonts::download_font(
    id = font_id,
    subsets = paste0(subsets, collapse = ","),
    variants = faces,
    output_dir = .font_dir
  )
  Sys.sleep(getOption("GFONTS_DOWNLOAD_SLEEPTIME"))
  gfonts::generate_css(
    prefer_local_source = FALSE,
    browser_support = "best",
    subsets = paste0(subsets, collapse = ","),
    id = font_id,
    variants = faces,
    font_dir = "../fonts/",
    output = css_file
  )

  # drop IE support as it generates issues - enough with IE8
  css_str <- readLines(css_file, encoding = "UTF-8")
  css_str <- css_str[!grepl("/* IE6-IE8 */", css_str, fixed = TRUE)]
  writeLines(css_str, css_file, useBytes = TRUE)
  TRUE
}


## gfonts_summary -----
gfonts_summary <- function() {
  all_gfonts
}

#' @useDynLib gdtools
#' @importFrom Rcpp sourceCpp
#' @importFrom grDevices is.raster
NULL

#' @title Version numbers of C libraries
#'
#' @description Return the runtime version of the Cairo and FreeType
#' libraries linked to the package.
#' @return An object of class \code{"numeric_version"}.
#' @examples
#' version_cairo()
#' version_freetype()
#' @export
version_cairo <- function() {
  ver <- version_cairo_()
  ver <- strsplit(ver, "\\.")[[1]]
  ver <- as.integer(ver)

  structure(list(ver), class = c("package_version", "numeric_version"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liberation.R
\name{register_liberationserif}
\alias{register_liberationserif}
\title{Register font 'Liberation Serif'}
\usage{
register_liberationserif(name = "Liberation Serif")
}
\arguments{
\item{name}{the name to use for the font family when
registering with 'systemfonts'. Using a custom name
(e.g. \code{"sans"}, \code{"serif"}, \code{"mono"}) allows devices like
'ragg' to resolve generic family names to this font.}
}
\value{
TRUE if the operation went ok.
}
\description{
Register font 'Liberation Serif' so that it can be used
with devices using the 'systemfonts' package, i.e. the 'flextable'
package and graphic outputs generated with the 'ragg', 'svglite'
and 'ggiraph' packages.
}
\examples{
register_liberationserif()
register_liberationserif(name = "serif")
}
\seealso{
\code{\link[=font_set_liberation]{font_set_liberation()}}, \code{\link[=font_set]{font_set()}}

Other functions for font management: 
\code{\link{addGFontHtmlDependency}()},
\code{\link{fonts_cache_dir}()},
\code{\link{gfontHtmlDependency}()},
\code{\link{install_gfont_script}()},
\code{\link{installed_gfonts}()},
\code{\link{liberationmonoHtmlDependency}()},
\code{\link{liberationsansHtmlDependency}()},
\code{\link{liberationserifHtmlDependency}()},
\code{\link{register_gfont}()},
\code{\link{register_liberationmono}()},
\code{\link{register_liberationsans}()}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_set.R
\name{font_liberation}
\alias{font_liberation}
\title{Liberation Font specification}
\usage{
font_liberation(variant = c("sans", "serif", "mono"))
}
\arguments{
\item{variant}{one of \code{"sans"}, \code{"serif"}, or \code{"mono"}.}
}
\value{
An object of class \code{font_spec}.
}
\description{
Create a font specification for a Liberation font,
to be used with \code{\link[=font_set]{font_set()}}.
}
\examples{
font_liberation("sans")
font_liberation("mono")
}
\seealso{
Other font set functions: 
\code{\link{font_google}()},
\code{\link{font_set}()},
\code{\link{font_set_auto}()},
\code{\link{font_set_liberation}()}
}
\concept{font set functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font-caching.R
\name{dummy_setup}
\alias{dummy_setup}
\title{Set up a temporary font cache}
\usage{
dummy_setup()
}
\value{
The path to the temporary cache directory (invisibly).
}
\description{
Create a temporary directory used as the
'Google Fonts' cache. This is intended for examples and tests.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fontconfig.R
\name{match_family}
\alias{match_family}
\title{Find best family match with systemfonts}
\usage{
match_family(font = "sans", bold = TRUE, italic = TRUE, debug = NULL)
}
\arguments{
\item{font}{family or face to match.}

\item{bold}{Whether to match a font featuring a \code{bold} face.}

\item{italic}{Whether to match a font featuring an \code{italic} face.}

\item{debug}{deprecated}
}
\description{
\code{match_family()} returns the best font family match.
}
\examples{
match_family("sans")
match_family("serif")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/str-metrics.R
\name{str_extents}
\alias{str_extents}
\title{Compute string extents.}
\usage{
str_extents(
  x,
  fontname = "sans",
  fontsize = 12,
  bold = FALSE,
  italic = FALSE,
  fontfile = ""
)
}
\arguments{
\item{x}{Character vector of strings to measure}

\item{fontname}{Font name}

\item{fontsize}{Font size}

\item{bold, italic}{Is text bold/italic?}

\item{fontfile}{Font file}
}
\description{
Determines the width and height of a bounding box that's big enough
to (just) enclose the provided text.
}
\examples{
str_extents(letters)
str_extents("Hello World!", bold = TRUE, italic = FALSE,
  fontname = "sans", fontsize = 12)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fonts.R
\name{gfontHtmlDependency}
\alias{gfontHtmlDependency}
\title{'Google Font' HTML dependency}
\usage{
gfontHtmlDependency(family = "Open Sans", subset = c("latin", "latin-ext"))
}
\arguments{
\item{family}{family name of a 'Google Fonts', for example, "Open Sans", "Roboto",
"Fira Code" or "Fira Sans Condensed". Complete list is available with the
following command:

\if{html}{\out{<div class="sourceCode">}}\preformatted{gfonts::get_all_fonts()$family |>
  unlist() |>
  unique() |>
  sort()
}\if{html}{\out{</div>}}}

\item{subset}{font subset, a character vector, it defaults to only "latin" and
"latin-ext" and can contains values such as "greek", "emoji", "chinese-traditional",

Run the following code to get a complete list:

\if{html}{\out{<div class="sourceCode">}}\preformatted{gfonts::get_all_fonts()$subsets |> unlist() |> unique() |> sort()
}\if{html}{\out{</div>}}}
}
\value{
an object defined with \code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}}.
}
\description{
Create an HTML dependency ready
to be used in 'Shiny' or 'R Markdown'.
}
\details{
It allows users to use fonts from 'Google Fonts' in an HTML page generated by 'shiny' or 'R Markdown'.
At the first request, the font files will be downloaded and stored in a cache on the
user's machine, thus avoiding many useless downloads or allowing to work with
these fonts afterwards without an Internet connection, in a docker image for example.
See \code{\link[=fonts_cache_dir]{fonts_cache_dir()}}.

The server delivering the font files should not be too busy. That's
why a one second pause is added after each download to respect the server's
limits. This time can be set with the option \code{GFONTS_DOWNLOAD_SLEEPTIME} which
must be a number of seconds.
}
\examples{
\dontrun{
if (check_gfonts()) {
  dummy_setup()
  gfontHtmlDependency(family = "Open Sans")
}
}
}
\seealso{
Other functions for font management: 
\code{\link{addGFontHtmlDependency}()},
\code{\link{fonts_cache_dir}()},
\code{\link{install_gfont_script}()},
\code{\link{installed_gfonts}()},
\code{\link{liberationmonoHtmlDependency}()},
\code{\link{liberationsansHtmlDependency}()},
\code{\link{liberationserifHtmlDependency}()},
\code{\link{register_gfont}()},
\code{\link{register_liberationmono}()},
\code{\link{register_liberationsans}()},
\code{\link{register_liberationserif}()}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/gdtools.R
\name{version_freetype}
\alias{version_freetype}
\alias{version_cairo}
\title{Version numbers of C libraries}
\usage{
version_freetype()

version_cairo()
}
\value{
An object of class \code{"numeric_version"}.
}
\description{
Return the runtime version of the Cairo and FreeType
libraries linked to the package.
}
\examples{
version_cairo()
version_freetype()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_set.R
\name{font_set_liberation}
\alias{font_set_liberation}
\title{Liberation font set}
\usage{
font_set_liberation()
}
\value{
An object of class \code{font_set}.
}
\description{
Shortcut to create a \code{\link[=font_set]{font_set()}} with all four
font roles using Liberation fonts (Sans, Serif, Mono) and
Liberation Sans as a fallback for symbols.
}
\examples{
fonts <- font_set_liberation()
fonts
}
\seealso{
\code{\link[=font_set]{font_set()}}, \code{\link[=font_set_auto]{font_set_auto()}}

Other font set functions: 
\code{\link{font_google}()},
\code{\link{font_liberation}()},
\code{\link{font_set}()},
\code{\link{font_set_auto}()}
}
\concept{font set functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fonts.R
\name{installed_gfonts}
\alias{installed_gfonts}
\title{List installed 'Google Fonts'}
\usage{
installed_gfonts()
}
\value{
families names as a character vector
}
\description{
List installed 'Google Fonts' that can be
found in the user cache directory.
}
\examples{
\dontrun{
if (check_gfonts()) {
  dummy_setup()
  register_gfont(family = "Roboto")
  installed_gfonts()
}
}
}
\seealso{
Other functions for font management:
\code{\link[=addGFontHtmlDependency]{addGFontHtmlDependency()}},
\code{\link[=fonts_cache_dir]{fonts_cache_dir()}},
\code{\link[=gfontHtmlDependency]{gfontHtmlDependency()}},
\code{\link[=install_gfont_script]{install_gfont_script()}},
\code{\link[=liberationmonoHtmlDependency]{liberationmonoHtmlDependency()}},
\code{\link[=liberationsansHtmlDependency]{liberationsansHtmlDependency()}},
\code{\link[=liberationserifHtmlDependency]{liberationserifHtmlDependency()}},
\code{\link[=register_gfont]{register_gfont()}},
\code{\link[=register_liberationmono]{register_liberationmono()}},
\code{\link[=register_liberationsans]{register_liberationsans()}},
\code{\link[=register_liberationserif]{register_liberationserif()}}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font-caching.R
\name{fonts_cache_dir}
\alias{fonts_cache_dir}
\alias{rm_fonts_cache}
\alias{init_fonts_cache}
\title{Manage Font Working Directory}
\usage{
fonts_cache_dir()

rm_fonts_cache()

init_fonts_cache()
}
\description{
Initialize or remove font directory used
to store downloaded font files.

This directory is managed by R function \code{\link[=R_user_dir]{R_user_dir()}} but can also
be defined in a non-user location by setting ENV variable \code{GDTOOLS_CACHE_DIR}
or by setting R option \code{GDTOOLS_CACHE_DIR}.

Its value can be read with the \code{fonts_cache_dir()} function.

The directory can be deleted with \code{rm_fonts_cache()} and
created with \code{init_fonts_cache()}.
}
\examples{
fonts_cache_dir()

options(GDTOOLS_CACHE_DIR = tempdir())
fonts_cache_dir()
options(GDTOOLS_CACHE_DIR = NULL)

Sys.setenv(GDTOOLS_CACHE_DIR = tempdir())
fonts_cache_dir()
Sys.setenv(GDTOOLS_CACHE_DIR = "")



init_fonts_cache()
dir.exists(fonts_cache_dir())

rm_fonts_cache()
dir.exists(fonts_cache_dir())
}
\seealso{
Other functions for font management: 
\code{\link{addGFontHtmlDependency}()},
\code{\link{gfontHtmlDependency}()},
\code{\link{install_gfont_script}()},
\code{\link{installed_gfonts}()},
\code{\link{liberationmonoHtmlDependency}()},
\code{\link{liberationsansHtmlDependency}()},
\code{\link{liberationserifHtmlDependency}()},
\code{\link{register_gfont}()},
\code{\link{register_liberationmono}()},
\code{\link{register_liberationsans}()},
\code{\link{register_liberationserif}()}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/str-metrics.R
\name{m_str_extents}
\alias{m_str_extents}
\title{Compute string extents for a vector of strings.}
\usage{
m_str_extents(
  x,
  fontname = "sans",
  fontsize = 10,
  bold = FALSE,
  italic = FALSE,
  fontfile = NULL
)
}
\arguments{
\item{x}{Character vector of strings to measure}

\item{fontname}{Font name. A character vector, recycled to match \code{x}.}

\item{fontsize}{Font size. A numeric vector, recycled to match \code{x}.}

\item{bold, italic}{Is text bold/italic? A logical vector, recycled to match \code{x}.}

\item{fontfile}{Font file. A character vector, recycled to match \code{x}.}
}
\description{
For each \code{x} element, determines the width and height of a bounding box that's big enough
to (just) enclose the provided text. Unit is points.
}
\note{
This function will be deprecated in a future release.
Use \code{\link[=strings_sizes]{strings_sizes()}} instead.
}
\examples{
\donttest{
# The first run can be slow when font caches are missing
# as font files are then being scanned to build those font caches.
m_str_extents(letters, fontsize = 1:26)
m_str_extents(letters[1:3],
  bold = c(TRUE, FALSE, TRUE),
  italic = c(FALSE, TRUE, TRUE),
  fontname = c("sans", "sans", "sans") )
}
}
\seealso{
Other functions for font metrics: 
\code{\link{str_metrics}()},
\code{\link{strings_sizes}()}
}
\concept{functions for font metrics}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_family_exists.R
\name{sys_fonts}
\alias{sys_fonts}
\title{List fonts for 'systemfonts'.}
\usage{
sys_fonts()
}
\value{
A data.frame of font information.
}
\description{
List system and registry fonts into a data.frame
containing columns such as path, family, style, weight and italic.
}
\examples{
sys_fonts()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{raster_write}
\alias{raster_write}
\title{Draw/preview a raster to a png file}
\usage{
raster_write(x, path, width = 480, height = 480, interpolate = FALSE)
}
\arguments{
\item{x}{A raster object}

\item{path}{name of the file to create}

\item{width, height}{Width and height in pixels.}

\item{interpolate}{A logical value indicating whether to linearly
interpolate the image.}
}
\description{
Draw/preview a raster to a png file
}
\examples{
r <- as.raster(matrix(hcl(0, 80, seq(50, 80, 10)),
 nrow = 4, ncol = 5))
filepng <- tempfile(fileext = ".png")
raster_write(x = r, path = filepng, width = 50, height = 50)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fonts.R
\name{check_gfonts}
\alias{check_gfonts}
\title{Checks the operability of 'gfonts'}
\usage{
check_gfonts(errors = FALSE)
}
\arguments{
\item{errors}{if TRUE, function is triggering errors if a condition
is not satisfied.}
}
\value{
TRUE or FALSE
}
\description{
Checks that 'gfonts' is installed and can be used.
Packages 'curl' and 'gfonts' must be installed and internet
connectivity must be active.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/str-metrics.R
\name{str_metrics}
\alias{str_metrics}
\title{Get font metrics for a string.}
\usage{
str_metrics(
  x,
  fontname = "sans",
  fontsize = 12,
  bold = FALSE,
  italic = FALSE,
  fontfile = ""
)
}
\arguments{
\item{x}{Character vector of strings to measure (only the first element is used).}

\item{fontname}{Font name.}

\item{fontsize}{Font size in points.}

\item{bold, italic}{Is text bold/italic?}

\item{fontfile}{Path to a font file. If empty, the font is
resolved by name via \code{systemfonts}.}
}
\value{
A named numeric vector with elements \code{width},
\code{ascent} and \code{descent} (unit is points).
}
\description{
Compute the width, ascent and descent of a single
string in points using Cairo text measurement.
}
\note{
This function will be deprecated in a future release.
Use \code{\link[=strings_sizes]{strings_sizes()}} instead.
}
\examples{
str_metrics("Hello World!")
}
\seealso{
Other functions for font metrics: 
\code{\link{m_str_extents}()},
\code{\link{strings_sizes}()}
}
\concept{functions for font metrics}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liberation-sans.R
\name{register_liberationsans}
\alias{register_liberationsans}
\title{Register font 'Liberation Sans'}
\usage{
register_liberationsans()
}
\value{
TRUE if the operation went ok.
}
\description{
Register font 'Liberation Sans' so that it can be used
with devices using the 'systemfonts' package, i.e. the 'flextable'
package and graphic outputs generated with the 'ragg', 'svglite'
and 'ggiraph' packages.
}
\seealso{
Other functions for font management: 
\code{\link{addGFontHtmlDependency}()},
\code{\link{fonts_cache_dir}()},
\code{\link{gfontHtmlDependency}()},
\code{\link{install_gfont_script}()},
\code{\link{installed_gfonts}()},
\code{\link{liberationsansHtmlDependency}()},
\code{\link{register_gfont}()}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fonts.R
\name{install_gfont_script}
\alias{install_gfont_script}
\title{Shell command to install a font from 'Google Fonts'}
\usage{
install_gfont_script(
  family = "Open Sans",
  subset = c("latin", "latin-ext"),
  platform = c("debian", "windows", "macos"),
  file = NULL
)
}
\arguments{
\item{family}{family name of a 'Google Fonts', for example, "Open Sans", "Roboto",
"Fira Code" or "Fira Sans Condensed". Complete list is available with the
following command:

\if{html}{\out{<div class="sourceCode">}}\preformatted{gfonts::get_all_fonts()$family |>
  unlist() |>
  unique() |>
  sort()
}\if{html}{\out{</div>}}}

\item{subset}{font subset, a character vector, it defaults to only "latin" and
"latin-ext" and can contains values such as "greek", "emoji", "chinese-traditional",

Run the following code to get a complete list:

\if{html}{\out{<div class="sourceCode">}}\preformatted{gfonts::get_all_fonts()$subsets |> unlist() |> unique() |> sort()
}\if{html}{\out{</div>}}}

\item{platform}{"debian" and "windows" and "macos" are supported.}

\item{file}{script file to generate, optional. If the parameter is
specified, a file will be generated ready for execution. If the
platform is Windows, administration rights are required to run
the script.}
}
\value{
the 'shell' or 'PowerShell' command as a string
}
\description{
Create a string containing a system command to execute
so that the font from 'Google Fonts' is installed on the system.
Its execution may require root permissions, in dockerfile for example.
}
\details{
It allows users to use fonts from 'Google Fonts' in an HTML page generated by 'shiny' or 'R Markdown'.
At the first request, the font files will be downloaded and stored in a cache on the
user's machine, thus avoiding many useless downloads or allowing to work with
these fonts afterwards without an Internet connection, in a docker image for example.
See \code{\link[=fonts_cache_dir]{fonts_cache_dir()}}.

The server delivering the font files should not be too busy. That's
why a one second pause is added after each download to respect the server's
limits. This time can be set with the option \code{GFONTS_DOWNLOAD_SLEEPTIME} which
must be a number of seconds.
}
\examples{
\dontrun{
if (check_gfonts()) {
  dummy_setup()
  install_gfont_script(family = "Roboto", platform = "macos")
}
}
}
\seealso{
Other functions for font management: 
\code{\link{addGFontHtmlDependency}()},
\code{\link{fonts_cache_dir}()},
\code{\link{gfontHtmlDependency}()},
\code{\link{installed_gfonts}()},
\code{\link{liberationsansHtmlDependency}()},
\code{\link{register_gfont}()},
\code{\link{register_liberationsans}()}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_set.R
\name{font_google}
\alias{font_google}
\title{Google Font specification}
\usage{
font_google(family, subset = c("latin", "latin-ext"))
}
\arguments{
\item{family}{family name of a 'Google Fonts', for example,
"Open Sans", "Roboto", "Fira Code" or "Fira Sans Condensed".}

\item{subset}{font subset, a character vector, it defaults to
only "latin" and "latin-ext".}
}
\value{
An object of class \code{font_spec}.
}
\description{
Create a font specification for a Google Font,
to be used with \code{\link[=font_set]{font_set()}}.
}
\examples{
font_google("Roboto")
}
\seealso{
Other font set functions: 
\code{\link{font_liberation}()},
\code{\link{font_set}()},
\code{\link{font_set_auto}()},
\code{\link{font_set_liberation}()}
}
\concept{font set functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/str-metrics.R
\name{glyphs_match}
\alias{glyphs_match}
\title{Validate glyph entries}
\usage{
glyphs_match(x, fontname = "sans", bold = FALSE, italic = FALSE, fontfile = "")
}
\arguments{
\item{x}{Character vector of strings}

\item{fontname}{Font name}

\item{bold, italic}{Is text bold/italic?}

\item{fontfile}{Font file}
}
\value{
a logical vector, if a character element is containing at least
a glyph that can not be matched in the font table, FALSE is returned.
}
\description{
Determines if strings contain glyphs not part of a font.
}
\examples{
glyphs_match(letters)
glyphs_match("\u265E", bold = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_set.R
\name{font_set}
\alias{font_set}
\title{Create a font set}
\usage{
font_set(sans = NULL, serif = NULL, mono = NULL, symbol = NULL)
}
\arguments{
\item{sans}{font for sans-serif text.}

\item{serif}{font for serif text.}

\item{mono}{font for monospace text.}

\item{symbol}{font for symbol text.}
}
\value{
An object of class \code{font_set} with elements:
\describe{
\item{sans}{character, the sans font family name (or NULL)}
\item{serif}{character, the serif font family name (or NULL)}
\item{mono}{character, the mono font family name (or NULL)}
\item{symbol}{character, the symbol font family name (or NULL)}
\item{dependencies}{list of \code{htmlDependency} objects}
\item{dsvg_fonts}{named list of family names for \code{dsvg()}}
\item{sources}{named list of source labels}
}
}
\description{
Bundle font registration, HTML dependencies, and
font family names into a single object. Each argument accepts
a \code{\link[=font_google]{font_google()}} specification, a \code{\link[=font_liberation]{font_liberation()}} specification,
or a plain character string naming a system font already available.

The returned object provides fields ready to use with
\code{\link[ggplot2:theme]{ggplot2::theme()}} (\verb{$sans}, \verb{$serif}, \verb{$mono}),
\code{girafe(fonts = ...)} or \code{dsvg(fonts = ...)} (\verb{$dsvg_fonts}),
and \code{girafe(dependencies = ...)} (\verb{$dependencies}).
}
\examples{
fonts <- font_set(sans = font_liberation("sans"))
fonts$sans
fonts$dsvg_fonts
fonts$dependencies
}
\seealso{
\code{\link[=font_set_liberation]{font_set_liberation()}}, \code{\link[=font_set_auto]{font_set_auto()}}

Other font set functions: 
\code{\link{font_google}()},
\code{\link{font_liberation}()},
\code{\link{font_set_auto}()},
\code{\link{font_set_liberation}()}
}
\concept{font set functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_set.R
\name{font_set_auto}
\alias{font_set_auto}
\title{Automatic font set}
\usage{
font_set_auto()
}
\value{
An object of class \code{font_set}.
}
\description{
Build a \code{\link[=font_set]{font_set()}} by detecting the best available
system fonts for each role (sans, serif, mono, symbol).

For each role a list of well-known fonts is tried in order.
When none is found on the system, the corresponding Liberation
font is used as a guaranteed offline fallback.\tabular{lll}{
   Role \tab Candidates (in order) \tab Fallback \cr
   sans \tab Arial, Helvetica, DejaVu Sans \tab Liberation Sans \cr
   serif \tab Times New Roman, Times, DejaVu Serif \tab Liberation Serif \cr
   mono \tab Courier New, Courier, DejaVu Sans Mono \tab Liberation Mono \cr
   symbol \tab Symbol, Apple Symbols \tab Liberation Sans \cr
}
}
\examples{
fonts <- font_set_auto()
fonts
}
\seealso{
\code{\link[=font_set]{font_set()}}, \code{\link[=font_set_liberation]{font_set_liberation()}}

Other font set functions: 
\code{\link{font_google}()},
\code{\link{font_liberation}()},
\code{\link{font_set}()},
\code{\link{font_set_liberation}()}
}
\concept{font set functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{raster_str}
\alias{raster_str}
\alias{raster_view}
\title{Draw/preview a raster into a string}
\usage{
raster_str(x, width = 480, height = 480, interpolate = FALSE)

raster_view(code)
}
\arguments{
\item{x}{A raster object}

\item{width, height}{Width and height in pixels.}

\item{interpolate}{A logical value indicating whether to linearly
interpolate the image.}

\item{code}{base64 code of a raster}
}
\description{
\code{raster_view} is a helper function for testing. It uses \code{htmltools}
to render a png as an image with base64 encoded data image.
}
\examples{
r <- as.raster(matrix(hcl(0, 80, seq(50, 80, 10)),
 nrow = 4, ncol = 5))
code <- raster_str(r, width = 50, height = 50)
if (interactive() && require("htmltools")) {
  raster_view(code = code)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fontconfig.R
\name{fontconfig_reinit}
\alias{fontconfig_reinit}
\title{reload Fontconfig configuration}
\usage{
fontconfig_reinit()
}
\description{
This function can be used to make fontconfig
reload font configuration files.
}
\note{
Fontconfig is not used anymore and that function will be deprecated
in the next release.
}
\author{
Paul Murrell
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liberation-sans.R
\name{liberationsansHtmlDependency}
\alias{liberationsansHtmlDependency}
\title{'Liberation Sans' Font HTML dependency}
\usage{
liberationsansHtmlDependency()
}
\description{
Create an HTML dependency ready
to be used in 'Shiny' or 'R Markdown' with
'Liberation Sans' Font.
}
\seealso{
\code{\link[=gfontHtmlDependency]{gfontHtmlDependency()}}

Other functions for font management: 
\code{\link{addGFontHtmlDependency}()},
\code{\link{fonts_cache_dir}()},
\code{\link{gfontHtmlDependency}()},
\code{\link{install_gfont_script}()},
\code{\link{installed_gfonts}()},
\code{\link{register_gfont}()},
\code{\link{register_liberationsans}()}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/str-metrics.R
\name{strings_sizes}
\alias{strings_sizes}
\title{Compute strings sizes}
\usage{
strings_sizes(
  x,
  fontname = "sans",
  fontsize = 10,
  bold = FALSE,
  italic = FALSE
)
}
\arguments{
\item{x}{A character vector of strings to measure. All arguments are vectorized
and recycled to match the length of \code{x}.}

\item{fontname}{A character vector specifying the font family name (e.g., "sans", "serif", "mono").
Default is "sans". This argument is vectorized.}

\item{fontsize}{A numeric vector specifying the font size in points.
Default is 10. This argument is vectorized.}

\item{bold}{A logical vector indicating whether the text should be bold.
Default is FALSE. This argument is vectorized.}

\item{italic}{A logical vector indicating whether the text should be italic.
Default is FALSE. This argument is vectorized.}
}
\description{
Determines widths, ascent and descent in inches.
Font lookup is performed by 'systemfonts' (so any font
registered via \code{\link[systemfonts:register_font]{systemfonts::register_font()}},
\code{\link[=register_gfont]{register_gfont()}}, or \code{\link[=font_set]{font_set()}} is found), then Cairo
computes the actual metrics. The results are accurate for
devices whose rendering finds the same font -- this is
guaranteed for 'systemfonts'-based devices (ragg, svglite,
ggiraph) and true for Cairo devices (\code{cairo_pdf()}, ...) when
the font is also installed at the system level. For devices
with their own font engine (\code{pdf()}, \code{png()}, ...) the
metrics may not match the rendering.
}
\examples{
strings_sizes(letters)
strings_sizes("Hello World!", bold = TRUE, italic = FALSE,
  fontname = "sans", fontsize = 12)
}
\seealso{
Other functions for font metrics: 
\code{\link{m_str_extents}()},
\code{\link{str_metrics}()}
}
\concept{functions for font metrics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liberation.R
\name{liberationserifHtmlDependency}
\alias{liberationserifHtmlDependency}
\title{'Liberation Serif' Font HTML dependency}
\usage{
liberationserifHtmlDependency()
}
\description{
Create an HTML dependency ready
to be used in 'Shiny' or 'R Markdown' with
'Liberation Serif' Font.
}
\seealso{
\code{\link[=font_set_liberation]{font_set_liberation()}}, \code{\link[=font_set]{font_set()}}, \code{\link[=gfontHtmlDependency]{gfontHtmlDependency()}}

Other functions for font management:
\code{\link[=addGFontHtmlDependency]{addGFontHtmlDependency()}},
\code{\link[=fonts_cache_dir]{fonts_cache_dir()}},
\code{\link[=gfontHtmlDependency]{gfontHtmlDependency()}},
\code{\link[=install_gfont_script]{install_gfont_script()}},
\code{\link[=installed_gfonts]{installed_gfonts()}},
\code{\link[=liberationmonoHtmlDependency]{liberationmonoHtmlDependency()}},
\code{\link[=liberationsansHtmlDependency]{liberationsansHtmlDependency()}},
\code{\link[=register_gfont]{register_gfont()}},
\code{\link[=register_liberationmono]{register_liberationmono()}},
\code{\link[=register_liberationsans]{register_liberationsans()}},
\code{\link[=register_liberationserif]{register_liberationserif()}}
}
\concept{functions for font management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/font_family_exists.R
\name{font_family_exists}
\alias{font_family_exists}
\title{Check if font family exists.}
\usage{
font_family_exists(font_family = "sans", system_only = FALSE)
}
\arguments{
\item{font_family}{font family name (case sensitive)}

\item{system_only}{If \code{TRUE}, only look in system-installed fonts
(as returned by \code{\link[systemfonts:system_fonts]{systemfonts::system_fonts()}}). This is useful
to verify that a font will be found by fontconfig, which is the
font resolution mechanism used by Cairo devices at rendering
time. When \code{FALSE} (the default), both system and registered
fonts are considered.}
}
\value{
A logical value
}
\description{
Check if a font family exists in available fonts.
}
\examples{
font_family_exists("sans")
font_family_exists("Arial")
font_family_exists("Courier")
font_family_exists("Arial", system_only = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fontconfig.R
\name{set_dummy_conf}
\alias{set_dummy_conf}
\title{Set and unset a minimalistic Fontconfig configuration}
\usage{
set_dummy_conf()
}
\description{
Set and unset a minimalistic Fontconfig configuration
}
\note{
Fontconfig is not used anymore and these functions will be deprecated
in the next release.
}
\keyword{internal}
